import json

import pytest
from monty.io import zopen

from emmet.core.mpid import AlphaID
from emmet.core.tasks import TaskDoc
from emmet.core.vasp.calc_types import RunType, TaskType, run_type, task_type
from emmet.core.vasp.validation import ValidationDoc
from emmet.core.vasp.validation_legacy import ValidationDoc as LegacyValidationDoc
from emmet.core.vasp.task_valid import TaskDocument
from emmet.core.testing_utils import DataArchive
from emmet.core.vasp.utils import discover_vasp_files

# As the validator itself is a separate package with its own tests,
# only check that its integration within emmet-core works.


def test_task_type():
    # TODO: Switch this to actual inputs?
    input_types = [
        ("NSCF Line", {"incar": {"ICHARG": 11}, "kpoints": {"labels": ["A"]}}),
        ("NSCF Uniform", {"incar": {"ICHARG": 11}}),
        ("Dielectric", {"incar": {"LEPSILON": True}}),
        ("DFPT Dielectric", {"incar": {"LEPSILON": True, "IBRION": 7}}),
        ("DFPT Dielectric", {"incar": {"LEPSILON": True, "IBRION": 8}}),
        ("DFPT", {"incar": {"IBRION": 7}}),
        ("DFPT", {"incar": {"IBRION": 8}}),
        ("Static", {"incar": {"NSW": 0}}),
    ]

    for _type, inputs in input_types:
        assert task_type(inputs) == TaskType(_type)


def test_run_type():
    params_sets = [
        ("GGA", {"GGA": "--"}),
        ("GGA+U", {"GGA": "--", "LDAU": True}),
        ("SCAN", {"METAGGA": "Scan"}),
        ("SCAN+U", {"METAGGA": "Scan", "LDAU": True}),
        ("r2SCAN", {"METAGGA": "R2SCAN"}),
        ("r2SCAN+U", {"METAGGA": "R2SCAN", "LDAU": True}),
        ("HFCus", {"LHFCALC": True}),
    ]

    for _type, params in params_sets:
        assert run_type(params) == RunType(_type)


@pytest.fixture(scope="session")
def tasks(test_dir):
    with zopen(test_dir / "test_si_tasks.json.gz") as f:
        data = json.load(f)

    return [TaskDoc(**d) for d in data]


def test_legacy_validator(tasks):
    validation_docs = [LegacyValidationDoc.from_task_doc(task) for task in tasks]

    assert len(validation_docs) == len(tasks)
    assert all([doc.valid for doc in validation_docs])


def test_legacy_validator_magmom(test_dir):
    # Task with Cr in structure - this is only element with MAGMOM check

    def cr_task_dict():
        # Model validation for TaskDoc serializes some fields.
        # Need to refresh reference unserialized dict to have
        # serialization for TaskDocument also work.
        with zopen(test_dir / "task_doc_mp-2766060.json.gz") as f:
            _cr_task_dict = json.load(f)
        return _cr_task_dict

    taskdoc = TaskDoc(**cr_task_dict())
    assert LegacyValidationDoc.from_task_doc(taskdoc).valid

    # test backwards compatibility
    taskdocument = TaskDocument(
        **{k: v for k, v in cr_task_dict().items() if k != "last_updated"}
    )
    assert LegacyValidationDoc.from_task_doc(taskdocument).valid

    # Change MAGMOM on Cr to fail magmom test
    td_bad_mag = TaskDoc(**cr_task_dict())
    td_bad_mag.calcs_reversed[0].output.outcar["magnetization"] = [
        {"tot": 6.0} if td_bad_mag.structure[ientry].species_string == "Cr" else entry
        for ientry, entry in enumerate(
            td_bad_mag.calcs_reversed[0].output.outcar["magnetization"]
        )
    ]
    assert not (valid_doc := LegacyValidationDoc.from_task_doc(td_bad_mag)).valid
    assert any("MAG" in repr(reason) for reason in valid_doc.reasons)

    # Remove magnetization tag to simulate spin-unpolarized (ISPIN = 1) calculation
    td_no_mag = TaskDoc(**cr_task_dict())
    del td_no_mag.calcs_reversed[0].output.outcar["magnetization"]
    assert LegacyValidationDoc.from_task_doc(td_no_mag).valid


def test_validator_failed_symmetry(test_dir):
    with zopen(test_dir / "failed_elastic_task.json.gz", "r") as f:
        failed_task = json.load(f)
    taskdoc = TaskDoc(**failed_task)
    validation = LegacyValidationDoc.from_task_doc(taskdoc)
    assert any("SYMMETRY" in repr(reason) for reason in validation.reasons)


def test_computed_entry(tasks):
    entries = [task.entry for task in tasks]
    ids = {e.entry_id for e in entries}
    assert ids == set(
        [
            AlphaID(id_str).string
            for id_str in {"mp-ddzda", "mp-dryyt", "mp-cmxxl", "mp-ft"}
        ]
    )


@pytest.fixture(scope="session")
def task_ldau(test_dir):
    with zopen(test_dir / "test_task.json.gz") as f:
        data = json.load(f)

    return TaskDoc(**data)


def test_ldau(task_ldau):
    task_ldau.input.is_hubbard = True
    assert task_ldau.run_type == RunType.GGA_U
    assert not LegacyValidationDoc.from_task_doc(task_ldau).valid


def test_ldau_validation(test_dir):
    with zopen(test_dir / "old_aflow_ggau_task.json.gz") as f:
        data = json.load(f)

    task = TaskDoc(**data)
    assert task.run_type == "GGA+U"

    valid = LegacyValidationDoc.from_task_doc(task)

    assert valid.valid


def test_validator_integration(test_dir):
    calc_dirs = {
        "Si_old_double_relax": ["VASP VERSION"],
        "Si_uniform": [],
        "Si_static": [],
    }
    for calc_dir, ref_reasons in calc_dirs.items():
        with DataArchive.extract(test_dir / "vasp" / f"{calc_dir}.json.gz") as tmp_dir:
            task_doc = TaskDoc.from_directory(tmp_dir)
        valid_doc = ValidationDoc.from_task_doc(task_doc, check_potcar=False)
        if len(ref_reasons) == 0:
            assert valid_doc.valid
        else:
            assert not valid_doc.valid
            assert all(
                any(reason_match in reason for reason in valid_doc.reasons)
                for reason_match in ref_reasons
            )


def test_from_file_meta_and_dir(test_dir):
    with DataArchive.extract(test_dir / "vasp" / "Si_static.json.gz") as tmp_dir:
        files_by_suffix = discover_vasp_files(tmp_dir)

        valid_doc_from_meta = ValidationDoc.from_file_metadata(
            files_by_suffix["standard"], check_potcar=False
        )

        valid_doc_from_dir = ValidationDoc.from_directory(tmp_dir, check_potcar=False)

    assert valid_doc_from_meta.valid

    for k in ValidationDoc.model_fields:
        if k in {"last_updated", "builder_meta"}:
            # these contain time-stamped fields that won't match
            continue
        assert getattr(valid_doc_from_meta, k) == getattr(valid_doc_from_dir, k)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from pydantic import Field
from pymatgen.analysis.graphs import StructureGraph
from pymatgen.analysis.local_env import NearNeighbors
from typing_extensions import TypedDict

from emmet.core.material_property import PropertyDoc
from emmet.core.types.pymatgen_types.structure_graph_adapter import StructureGraphType

AVAILABLE_METHODS = {nn.__name__: nn for nn in NearNeighbors.__subclasses__()}

if TYPE_CHECKING:

    from pymatgen.core import Structure
    from typing_extensions import Self

    from emmet.core.mpid import MPID


class TypedBondLengthStatsDict(TypedDict):
    all_weights: list[float]
    min: float
    max: float
    mean: float
    variance: float


class BondingDoc(PropertyDoc):
    """Structure graphs representing chemical bonds calculated from structure
    using near neighbor strategies as defined in pymatgen."""

    property_name: str = "bonding"

    structure_graph: StructureGraphType = Field(
        description="Structure graph",
    )

    method: str = Field(description="Method used to compute structure graph.")

    bond_types: dict[str, list[float]] = Field(
        description="Dictionary of bond types to their length, e.g. a Fe-O to "
        "a list of the lengths of Fe-O bonds in Angstrom."
    )
    bond_length_stats: TypedBondLengthStatsDict = Field(
        description="Dictionary of statistics of bonds in structure "
    )
    coordination_envs: list[str] = Field(
        description="List of co-ordination environments, e.g. ['Mo-S(6)', 'S-Mo(3)']."
    )
    coordination_envs_anonymous: list[str] = Field(
        description="List of co-ordination environments without elements "
        "present, e.g. ['A-B(6)', 'A-B(3)']."
    )

    @classmethod
    def from_structure(
        cls,
        structure: Structure,
        material_id: str | MPID,
        preferred_methods: tuple[str | NearNeighbors, ...] = (
            "CrystalNN",
            "MinimumDistanceNN",
        ),
        **kwargs,
    ) -> Self | None:
        """
        Calculate

        :param structure: ideally an oxidation state-decorated structure
        :param material_id: mpid
        :param preferred_methods: list of strings of NearNeighbor classes or NearNeighbor instances
        :param deprecated: whether source material is or is not deprecated
        :param kwargs: to pass to PropertyDoc
        :return:
        """

        bonding_info = None
        bond_analysis_methods: list[NearNeighbors] = [
            AVAILABLE_METHODS[method]() if isinstance(method, str) else method
            for method in preferred_methods
        ]

        warnings: list[str] = []

        for method in bond_analysis_methods:
            try:
                sg = StructureGraph.from_local_env_strategy(structure, method)

                # ensure edge weights are specifically bond lengths
                edge_weights = []
                for u, v, d in sg.graph.edges(data=True):
                    jimage = np.array(d["to_jimage"])
                    dist = sg.structure.get_distance(u, v, jimage=jimage)
                    edge_weights.append((u, v, d["to_jimage"], dist))
                for u, v, to_jimage, dist in edge_weights:
                    sg.alter_edge(u, v, to_jimage=to_jimage, new_weight=dist)

                bonding_info = {
                    "method": method.__class__.__name__,
                    "structure_graph": sg,
                    "bond_types": sg.types_and_weights_of_connections,
                    "bond_length_stats": sg.weight_statistics,
                    "coordination_envs": sg.types_of_coordination_environments(),
                    "coordination_envs_anonymous": sg.types_of_coordination_environments(
                        anonymous=True
                    ),
                }

                break

            except Exception as e:
                warnings.append(
                    f"Failed to calculate bonding: {material_id} {method} {e}"
                )

        if bonding_info:
            return super().from_structure(
                meta_structure=structure,
                material_id=material_id,
                warnings=warnings,
                **bonding_info,
                **kwargs,
            )
        return None



from __future__ import annotations

import logging
from collections import defaultdict
from typing import TYPE_CHECKING

import numpy as np
from pydantic import Field
from pymatgen.analysis.bond_valence import BVAnalyzer
from pymatgen.core import Structure
from pymatgen.core.periodic_table import Specie

from emmet.core.material_property import PropertyDoc
from emmet.core.types.enums import ValueEnum
from emmet.core.types.pymatgen_types.structure_adapter import StructureType

if TYPE_CHECKING:
    from emmet.core.types.typing import IdentifierType


class OxiStateAssigner(ValueEnum):

    MANUAL = "Manual"
    BVA = "Bond Valence Analysis"
    GUESS = "Oxidation State Guess"


class OxidationStateDoc(PropertyDoc):
    """Oxidation states computed from the structure"""

    property_name: str = "oxidation"

    structure: StructureType = Field(
        ...,
        description="The structure used in the generation of the oxidation state data.",
    )
    possible_species: list[str] = Field(
        description="Possible charged species in this material."
    )
    possible_valences: list[float] = Field(
        description="List of valences for each site in this material."
    )
    average_oxidation_states: dict[str, float] = Field(
        description="Average oxidation states for each unique species."
    )
    method: OxiStateAssigner | None = Field(
        None, description="Method used to compute oxidation states."
    )

    @classmethod
    def from_structure(
        cls, structure: Structure, material_id: IdentifierType | None = None, **kwargs
    ):

        # Check if structure already has oxidation states,
        # if so pass this along unchanged with "method" == "manualx"
        struct_valences: list[float | None] = []
        species = []

        method = None
        if _method := kwargs.pop("method", None):
            method = _method.lower()

        site_oxidation_list = defaultdict(list)
        for site in structure:
            if (oxi_state := getattr(site.species, "oxi_state", None)) and hasattr(
                site.species, "element"
            ):
                site_oxidation_list[site.species.element].append(oxi_state)
                species.append(site.species)
            struct_valences.append(oxi_state)

        average_oxidation_states: dict[str, float] = {
            str(el): np.mean(oxi_states)
            for el, oxi_states in site_oxidation_list.items()
        }

        if any(struct_valences) and (not method):
            d = {
                "possible_species": species,
                "possible_valences": struct_valences,
                "average_oxidation_states": average_oxidation_states,
                "method": OxiStateAssigner.MANUAL,
                "state": "successful",
            }
            return super().from_structure(
                meta_structure=structure, material_id=material_id, **d, **kwargs
            )

        # otherwise, continue with assignment

        structure.remove_oxidation_states()

        # Null document
        d = {
            "possible_species": [],
            "possible_valences": [],
            "average_oxidation_states": {},
            "method": method or OxiStateAssigner.BVA,
            "material_id": material_id,
        }

        if d["method"] == OxiStateAssigner.BVA:
            try:
                bva = BVAnalyzer()
                valences = bva.get_valences(structure)
                possible_species = {
                    str(Specie(structure[idx].specie, oxidation_state=valence))
                    for idx, valence in enumerate(valences)
                }

                structure.add_oxidation_state_by_site(valences)

                # construct a dict of average oxi_states for use
                # by MP2020 corrections. The format should mirror
                # the output of the first element from Composition.oxi_state_guesses()
                # e.g. {'Li': 1.0, 'O': -2.0}

                site_oxidation_list = defaultdict(list)
                for site in structure:
                    site_oxidation_list[site.specie.element].append(
                        site.specie.oxi_state
                    )

                oxi_state_dict = {
                    str(el): np.mean(oxi_states) for el, oxi_states in site_oxidation_list.items()  # type: ignore
                }

                d.update(
                    possible_species=list(possible_species),
                    possible_valences=valences,
                    average_oxidation_states=oxi_state_dict,
                )

            except Exception:
                logging.debug(
                    f"BVAnalyzer failed for {structure.composition.reduced_composition}. Trying oxi_state_guesses."
                )
                d["method"] = OxiStateAssigner.GUESS

        if d["method"] == OxiStateAssigner.GUESS:
            try:
                first_oxi_state_guess = structure.composition.oxi_state_guesses(
                    max_sites=-50
                )[0]
                valences = [
                    first_oxi_state_guess[site.species_string] for site in structure
                ]
                possible_species = {
                    str(Specie(el, oxidation_state=valence))
                    for el, valence in first_oxi_state_guess.items()
                }

                structure.add_oxidation_state_by_site(valences)

                d.update(
                    possible_species=list(possible_species),
                    possible_valences=valences,
                    average_oxidation_states=first_oxi_state_guess,
                )

            except Exception as e:
                logging.error("Oxidation state guess failed with: {}".format(e))
                d["warnings"] = ["Oxidation state guessing failed."]
                d["state"] = "unsuccessful"
                d["method"] = None

        return super().from_structure(
            meta_structure=structure,
            **d,
            **kwargs,
        )



"""Core definition for Polar property Document"""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from pydantic import BaseModel, Field
from pymatgen.analysis.piezo import PiezoTensor as BasePiezoTensor
from pymatgen.core.tensors import Tensor

from emmet.core.material_property import PropertyDoc
from emmet.core.math import Matrix3D
from emmet.core.settings import EmmetSettings
from emmet.core.types.pymatgen_types.ir_dielectric_tensor_adapter import (
    IRDielectricTensorType,
)

if TYPE_CHECKING:
    from pymatgen.core.structure import Structure

    from emmet.core.types.typing import IdentifierType


SETTINGS = EmmetSettings()

Vector = list[float]
PiezoTensor = list[Vector]


class DielectricDoc(PropertyDoc):
    """
    A dielectric property block
    """

    property_name: str = "dielectric"

    total: Matrix3D = Field(description="Total dielectric tensor.")
    ionic: Matrix3D = Field(description="Ionic contribution to dielectric tensor.")
    electronic: Matrix3D = Field(
        description="Electronic contribution to dielectric tensor."
    )

    e_total: float = Field(description="Total electric permittivity.")
    e_ionic: float = Field(
        description="Electric permittivity from atomic rearrangement."
    )
    e_electronic: float = Field(
        description="Electric permittivity due to electrons rearrangement."
    )

    n: float = Field(description="Refractive index.")

    @classmethod
    def from_ionic_and_electronic(
        cls,
        ionic: Matrix3D,
        electronic: Matrix3D,
        structure: Structure,
        material_id: IdentifierType | None = None,
        **kwargs,
    ):
        ionic_tensor = Tensor(ionic).convert_to_ieee(structure)
        electronic_tensor = Tensor(electronic).convert_to_ieee(structure)

        total = ionic_tensor + electronic_tensor

        return super().from_structure(
            meta_structure=structure,
            material_id=material_id,
            **{
                "total": total.tolist(),
                "ionic": ionic_tensor.tolist(),
                "electronic": electronic_tensor.tolist(),
                "e_total": np.average(np.diagonal(total)),
                "e_ionic": np.average(np.diagonal(ionic_tensor)),
                "e_electronic": np.average(np.diagonal(electronic_tensor)),
                "n": np.sqrt(np.average(np.diagonal(electronic_tensor))),
            },
            **kwargs,
        )


class PiezoelectricDoc(PropertyDoc):
    """
    A dielectric package block
    """

    property_name: str = "piezoelectric"

    total: PiezoTensor = Field(description="Total piezoelectric tensor in C/m²")
    ionic: PiezoTensor = Field(
        description="Ionic contribution to piezoelectric tensor in C/m²"
    )
    electronic: PiezoTensor = Field(
        description="Electronic contribution to piezoelectric tensor in C/m²"
    )

    e_ij_max: float = Field(description="Piezoelectric modulus")
    max_direction: list[int] = Field(
        description="Miller direction for maximum piezo response"
    )
    strain_for_max: list[float] = Field(
        description="Normalized strain direction for maximum piezo repsonse"
    )

    @classmethod
    def from_ionic_and_electronic(
        cls,
        ionic: PiezoTensor,
        electronic: PiezoTensor,
        structure: Structure,
        material_id: IdentifierType | None = None,
        **kwargs,
    ):
        ionic_tensor = BasePiezoTensor.from_vasp_voigt(ionic)
        electronic_tensor = BasePiezoTensor.from_vasp_voigt(electronic)
        total: BasePiezoTensor = ionic_tensor + electronic_tensor  # type: ignore[assignment]

        # Symmeterize Convert to IEEE orientation
        total = total.convert_to_ieee(structure)
        ionic_tensor = ionic_tensor.convert_to_ieee(structure)
        electronic_tensor = electronic_tensor.convert_to_ieee(structure)

        directions, charges, strains = np.linalg.svd(total.voigt, full_matrices=False)
        max_index = np.argmax(np.abs(charges))

        max_direction = directions[max_index]

        # Allow a max miller index of 10
        min_val = np.abs(max_direction)
        min_val = min_val[min_val > (np.max(min_val) / SETTINGS.MAX_PIEZO_MILLER)]
        min_val = np.min(min_val)

        return super().from_structure(
            meta_structure=structure,
            material_id=material_id,
            **{
                "total": total.zeroed().voigt.tolist(),
                "ionic": ionic_tensor.zeroed().voigt.tolist(),
                "electronic": electronic_tensor.zeroed().voigt.tolist(),
                "e_ij_max": charges[max_index],
                "max_direction": tuple(np.round(max_direction / min_val)),
                "strain_for_max": tuple(strains[max_index]),
            },
            **kwargs,
        )


class BornEffectiveCharges(BaseModel):
    """
    A block for the Born effective charges
    """

    value: list[Matrix3D] | None = Field(
        None, description="Value of the Born effective charges."
    )

    symmetrized_value: list[Matrix3D] | None = Field(
        None,
        description="Value of the Born effective charges after symmetrization to obey the"
        "charge neutrality sum rule.",
    )

    cnsr_break: float | None = Field(
        None,
        description="The maximum breaking of the charge neutrality sum "
        "rule (CNSR) in the Born effective charges.",
    )


class IRDielectric(BaseModel):
    """
    A block for the pymatgen IRDielectricTensor object
    """

    ir_dielectric_tensor: IRDielectricTensorType | None = Field(
        None, description="Serialized version of a pymatgen IRDielectricTensor object."
    )



"""Current MP tools to validate VASP calculations."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pydantic import Field
from pymatgen.io.validation.common import (
    LightOutcar,
    LightVasprun,
    PotcarSummaryStats,
    VaspFiles,
    VaspInputSafe,
)
from pymatgen.io.validation.validation import REQUIRED_VASP_FILES, VaspValidator
from pymatgen.io.vasp import Incar

from emmet.core.base import EmmetBaseModel
from emmet.core.types.typing import DateTimeType, IdentifierType
from emmet.core.utils import arrow_incompatible
from emmet.core.vasp.calc_types.enums import CalcType, RunType
from emmet.core.vasp.calculation import Calculation
from emmet.core.vasp.task_valid import TaskDocument
from emmet.core.vasp.utils import FileMetadata, discover_vasp_files

if TYPE_CHECKING:
    from pathlib import Path

    from typing_extensions import Self

    from emmet.core.tasks import TaskDoc


@arrow_incompatible
class ValidationDoc(EmmetBaseModel, VaspValidator):
    """
    Validation document for a VASP calculation
    """

    task_id: IdentifierType | None = Field(
        None, description="The task_id for this validation document"
    )

    last_updated: DateTimeType = Field(
        description="The most recent time when this document was updated.",
    )
    nelements: int | None = Field(None, description="Number of elements.")
    symmetry_number: int | None = Field(
        None,
        title="Space Group Number",
        description="The spacegroup number for the lattice.",
    )
    run_type: RunType | None = Field(
        None, description="The run type of the calculation"
    )
    calc_type: CalcType | None = Field(None, description="The calculation type.")

    @classmethod
    def from_file_metadata(cls, file_meta: list[FileMetadata], **kwargs) -> Self:
        """Validate files from a list of their metadata."""
        vasp_file_paths: dict[str, Path] = {}
        for f in file_meta:
            if (
                len(matched_files := [vf for vf in REQUIRED_VASP_FILES if vf in f.name])
                > 0
            ):
                vasp_file_paths[matched_files[0].lower().split(".")[0]] = f.path
        return cls.from_vasp_input(vasp_file_paths=vasp_file_paths, **kwargs)

    @staticmethod
    def task_doc_to_vasp_files(task_doc: TaskDoc | TaskDocument) -> VaspFiles:
        """Convert an emmet.core TaskDoc or legacy TaskDocument to VaspFiles."""

        if isinstance(task_doc, TaskDocument):
            final_calc = Calculation(**task_doc.calcs_reversed[0])
        else:
            final_calc = task_doc.calcs_reversed[0]

        potcar_stats = None
        if final_calc.input.potcar_spec:
            potcar_stats = [
                PotcarSummaryStats(
                    titel=ps.titel,
                    keywords=ps.summary_stats["keywords"] if ps.summary_stats else None,
                    stats=ps.summary_stats["stats"] if ps.summary_stats else None,
                    lexch=(
                        "pe" if final_calc.input.potcar_type[0] == "PAW_PBE" else "ca"
                    ),
                )
                for ps in final_calc.input.potcar_spec
            ]

        # Issue with legacy data: VASP version can include date info - remove here
        vasp_version = None
        if len(split_vasp_ver := final_calc.vasp_version.split(".")) > 0:
            vasp_version = ".".join(split_vasp_ver[: min(3, len(split_vasp_ver))])

        return VaspFiles(
            user_input=VaspInputSafe(  # type: ignore[call-arg]
                incar=Incar(final_calc.input.incar),
                kpoints=final_calc.input.kpoints,
                structure=final_calc.input.structure,
                potcar=potcar_stats,
            ),
            outcar=LightOutcar(
                **{
                    k: final_calc.output.outcar.get(k)
                    for k in ("drift", "magnetization")
                }
            ),
            vasprun=LightVasprun(  # type: ignore[call-arg]
                vasp_version=vasp_version,  # type: ignore[arg-type]
                ionic_steps=[
                    ionic_step.model_dump()
                    for ionic_step in final_calc.output.ionic_steps
                ],
                final_energy=task_doc.output.energy,
                final_structure=task_doc.output.structure,
                kpoints=final_calc.input.kpoints,
                parameters=final_calc.input.parameters,
                bandgap=final_calc.output.bandgap,
            ),
        )

    @classmethod
    def from_task_doc(cls, task_doc: TaskDoc | TaskDocument, **kwargs) -> Self:
        """Validate a VASP calculation represented by an emmet.core TaskDoc/ument."""
        vasp_files = cls.task_doc_to_vasp_files(task_doc)

        for k in ("run_type", "calc_type"):
            if not kwargs.get(k):
                kwargs[k] = getattr(task_doc, k, None)

        if not kwargs.get("symmetry_number") and task_doc.symmetry:
            kwargs["symmetry_number"] = task_doc.symmetry.number

        return cls.from_vasp_input(vasp_files=vasp_files, **kwargs)

    @classmethod
    def from_directory(cls, dir_name: str | Path, **kwargs) -> Self:
        """Override parent model to use file discovery method."""
        vasp_files = discover_vasp_files(dir_name)

        # NB: this will pick "standard" over "relax*" if present,
        # and will select the last "relax*" if those are the only
        # types present
        final_calc_name = sorted(vasp_files)[-1]
        return cls.from_file_metadata(vasp_files[final_calc_name], **kwargs)



from pydantic import BaseModel, Field

from emmet.core.types.enums import ValueEnum
from emmet.core.types.typing import DateTimeType


class MessageType(ValueEnum):
    generic = "generic"
    warning = "warning"


class MessagesDoc(BaseModel):
    """
    Defines data for user messages
    """

    title: str | None = Field(
        None,
        title="Title",
        description="Generic title or short summary for the message.",
    )

    body: str | None = Field(
        None, title="Body", description="Main text body of message."
    )

    authors: list[str] | None = Field(
        None,
        title="Title",
        description="Generic title or short summary for the message.",
    )

    type: MessageType = Field(  # type: ignore[assignment]
        MessageType.generic,
        title="Type",
        description="The type of the message.",
    )

    last_updated: DateTimeType = Field(
        title="Last Updated",
        description="The last updated UTC timestamp for the message.",
    )



from emmet.core.synthesis.core import (
    SynthesisRecipe,
    SynthesisTypeEnum,
    SynthesisSearchResultModel,
)
from emmet.core.synthesis.materials import (
    Component,
    ExtractedMaterial,
)
from emmet.core.synthesis.operations import (
    Value,
    Conditions,
    Operation,
    OperationTypeEnum,
)
from emmet.core.synthesis.reaction import (
    FormulaPart,
    ReactionFormula,
)



from __future__ import annotations

import re
from hashlib import blake2b
from typing import TYPE_CHECKING

from monty.json import MSONable
from pydantic import Field

from emmet.core.molecules import MolPropertyOrigin
from emmet.core.molecules.molecule_property import PropertyDoc
from emmet.core.mpid import MPculeID
from emmet.core.qchem.task import TaskDocument
from emmet.core.utils import set_msonable_type_adapter

if TYPE_CHECKING:
    from typing import Any

__author__ = "Evan Spotte-Smith <ewcspottesmith@lbl.gov>"


@set_msonable_type_adapter
class NaturalPopulation(MSONable):
    def __init__(
        self,
        atom_index: int,
        core_electrons: float,
        valence_electrons: float,
        rydberg_electrons: float,
        total_electrons: float,
    ):
        """
        Basic description of an atomic electron population.

        :param atom_index (int):
        :param core_electrons (float): Number of core electrons on this atom
        :param valence_electrons (float): Number of valence electrons on this atom
        :param rydberg_electrons (float): Number of Rydberg electrons on this atom
        :param total_electrons (float): Total number of electrons on this atom
        """

        self.atom_index = int(atom_index)
        self.core_electrons = float(core_electrons)
        self.valence_electrons = float(valence_electrons)
        self.rydberg_electrons = float(rydberg_electrons)
        self.total_electrons = float(total_electrons)


@set_msonable_type_adapter
class LonePair(MSONable):
    def __init__(
        self,
        index: int,
        number: int,
        atom_index: int,
        s_character: float,
        p_character: float,
        d_character: float,
        f_character: float,
        occupancy: float,
        type_code: str,
    ):
        """
        Basic description of a lone pair (LP) natural bonding orbital.

        :param index (int): Lone pair index from NBO. 1-indexed
        :param number (int): Another index, for cases where there are multiple lone
            pairs on an atom. 1-indexed
        :param atom_index (int): 0-indexed
        :param s_character (float): What fraction of this orbital is s-like in nature.
        :param p_character (float): What fraction of this orbital is p-like in nature.
        :param d_character (float): What fraction of this orbital is d-like in nature.
        :param f_character (float): What fraction of this orbital is f-like in nature.
        :param occupancy (float): Total electron occupancy of this lone pair
        :param type_code (str): Description of this lone pair (ex: "LV" for "lone valence")
        """

        self.index = int(index)
        self.number = int(number)
        self.atom_index = int(atom_index)
        self.s_character = float(s_character)
        self.p_character = float(p_character)
        self.d_character = float(d_character)
        self.f_character = float(f_character)
        self.occupancy = float(occupancy)
        self.type_code = type_code


@set_msonable_type_adapter
class Bond(MSONable):
    def __init__(
        self,
        index: int,
        number: int,
        atom1_index: int,
        atom2_index: int,
        atom1_s_character: float,
        atom2_s_character: float,
        atom1_p_character: float,
        atom2_p_character: float,
        atom1_d_character: float,
        atom2_d_character: float,
        atom1_f_character: float,
        atom2_f_character: float,
        atom1_polarization: float,
        atom2_polarization: float,
        atom1_polarization_coeff: float,
        atom2_polarization_coeff: float,
        occupancy: float,
        type_code: str,
    ):
        """
        Basic description of a bond (BD) natural bonding orbital.

        :param index: Bond orbital index from NBO. 1-indexed.
        :param number: Another index, for cases where there are multiple bonds
            between two atoms. 1-indexed
        :param atom1_index: Index of first atom involved in this orbital. 0-indexed
        :param atom2_index: Index of second atom involved in this orbital. 0-indexed
        :param atom1_s_character: What fraction of this orbital comes from atom 1 s electrons
        :param atom2_s_character: What fraction of this orbital comes from atom 2 s electrons
        :param atom1_p_character: What fraction of this orbital comes from atom 1 p electrons
        :param atom2_p_character: What fraction of this orbital comes from atom 2 p electrons
        :param atom1_d_character: What fraction of this orbital comes from atom 1 d electrons
        :param atom2_d_character: What fraction of this orbital comes from atom 2 d electrons
        :param atom1_f_character: What fraction of this orbital comes from atom 1 f electrons
        :param atom2_f_character: What fraction of this orbital comes from atom 2 f electrons
        :param atom1_polarization: Percentage of polarization from atom 1
        :param atom2_polarization: Percentage of polarization from atom 2
        :param atom1_polarization_coeff: Polarization coefficient of atom 1
        :param atom2_polarization_coeff: Polarization coefficient of atom 2
        :param occupancy: Total electron occupancy of this orbital
        :param type_code: Description of this bonding orbital (ex: BD for bonding,
            BD* for anti-bonding)
        """

        self.index = int(index)
        self.number = int(number)
        self.atom1_index = int(atom1_index)
        self.atom2_index = int(atom2_index)
        self.atom1_s_character = float(atom1_s_character)
        self.atom2_s_character = float(atom2_s_character)
        self.atom1_p_character = float(atom1_p_character)
        self.atom2_p_character = float(atom2_p_character)
        self.atom1_d_character = float(atom1_d_character)
        self.atom2_d_character = float(atom2_d_character)
        self.atom1_f_character = float(atom1_f_character)
        self.atom2_f_character = float(atom2_f_character)
        self.atom1_polarization = float(atom1_polarization)
        self.atom2_polarization = float(atom2_polarization)
        self.atom1_polarization_coeff = float(atom1_polarization_coeff)
        self.atom2_polarization_coeff = float(atom2_polarization_coeff)
        self.occupancy = float(occupancy)
        self.type_code = type_code


@set_msonable_type_adapter
class ThreeCenterBond(MSONable):
    def __init__(
        self,
        index: int,
        number: int,
        atom1_index: int,
        atom2_index: int,
        atom3_index: int,
        atom1_s_character: float,
        atom2_s_character: float,
        atom3_s_character: float,
        atom1_p_character: float,
        atom2_p_character: float,
        atom3_p_character: float,
        atom1_d_character: float,
        atom2_d_character: float,
        atom3_d_character: float,
        atom1_f_character: float,
        atom2_f_character: float,
        atom3_f_character: float,
        atom1_polarization: float,
        atom2_polarization: float,
        atom3_polarization: float,
        atom1_polarization_coeff: float,
        atom2_polarization_coeff: float,
        atom3_polarization_coeff: float,
        occupancy: float,
        type_code: str,
    ):
        """
        Basic description of a three-center bond (3C) natural bonding orbital.

        :param index: Bond orbital index from NBO. 1-indexed.
        :param number: Another index, for cases where there are multiple bonds
            between two atoms. 1-indexed
        :param atom1_index: Index of first atom involved in this orbital. 0-indexed
        :param atom2_index: Index of second atom involved in this orbital. 0-indexed
        :param atom3_index: Index of third atom involved in this orbital. 0-indexed
        :param atom1_s_character: What fraction of this orbital comes from atom 1 s electrons
        :param atom2_s_character: What fraction of this orbital comes from atom 2 s electrons
        :param atom3_s_character: What fraction of this orbital comes from atom 3 s electrons
        :param atom1_p_character: What fraction of this orbital comes from atom 1 p electrons
        :param atom2_p_character: What fraction of this orbital comes from atom 2 p electrons
        :param atom3_p_character: What fraction of this orbital comes from atom 3 p electrons
        :param atom1_d_character: What fraction of this orbital comes from atom 1 d electrons
        :param atom2_d_character: What fraction of this orbital comes from atom 2 d electrons
        :param atom3_d_character: What fraction of this orbital comes from atom 3 d electrons
        :param atom1_f_character: What fraction of this orbital comes from atom 1 f electrons
        :param atom2_f_character: What fraction of this orbital comes from atom 2 f electrons
        :param atom3_f_character: What fraction of this orbital comes from atom 3 f electrons
        :param atom1_polarization: Percentage of polarization from atom 1
        :param atom2_polarization: Percentage of polarization from atom 2
        :param atom3_polarization: Percentage of polarization from atom 3
        :param atom1_polarization_coeff: Polarization coefficient of atom 1
        :param atom2_polarization_coeff: Polarization coefficient of atom 2
        :param atom3_polarization_coeff: Polarization coefficient of atom 3
        :param occupancy: Total electron occupancy of this orbital
        :param type_code: Description of this bonding orbital (ex: 3C for three-center bond,
            3C* for three-center anti-bond)
        """

        self.index = int(index)
        self.number = int(number)
        self.atom1_index = int(atom1_index)
        self.atom2_index = int(atom2_index)
        self.atom3_index = int(atom3_index)
        self.atom1_s_character = float(atom1_s_character)
        self.atom2_s_character = float(atom2_s_character)
        self.atom3_s_character = float(atom3_s_character)
        self.atom1_p_character = float(atom1_p_character)
        self.atom2_p_character = float(atom2_p_character)
        self.atom3_p_character = float(atom3_p_character)
        self.atom1_d_character = float(atom1_d_character)
        self.atom2_d_character = float(atom2_d_character)
        self.atom3_d_character = float(atom3_d_character)
        self.atom1_f_character = float(atom1_f_character)
        self.atom2_f_character = float(atom2_f_character)
        self.atom3_f_character = float(atom3_f_character)
        self.atom1_polarization = float(atom1_polarization)
        self.atom2_polarization = float(atom2_polarization)
        self.atom3_polarization = float(atom3_polarization)
        self.atom1_polarization_coeff = float(atom1_polarization_coeff)
        self.atom2_polarization_coeff = float(atom2_polarization_coeff)
        self.atom3_polarization_coeff = float(atom3_polarization_coeff)
        self.occupancy = float(occupancy)
        self.type_code = type_code


@set_msonable_type_adapter
class Hyperbond(MSONable):
    def __init__(
        self,
        index: int,
        atom1_index: int,
        atom2_index: int,
        atom3_index: int,
        fraction_12: float,
        fraction_23: float,
        occupancy: float,
        bond_index_12: int,
        lp_index_3: int,
        hybrid_index_1: int,
        hybrid_index_2: int,
        hybrid_index_3: int,
    ):
        """
        Basic description of a three-center, four-electron hyperbond (3CHB).

        :param index: Bond orbital index from NBO. 1-indexed.
        :param atom1_index: Index of first atom involved in this orbital. 0-indexed
        :param atom2_index: Index of second atom involved in this orbital. 0-indexed
        :param atom3_index: Index of third atom involved in this orbital. 0-indexed
        :param fraction_12: What fraction of this hyperbond comes from the bond between atoms 1 and 2
        :param fraction_23: What fraction of this hyperbond comes from the bond between atoms 2 and 3
        :param occupancy: Total electron occupancy of this orbital
        :param bond_index_12: Index of the bond orbital between atoms used 1 and 2 used to make this hyperbond
        :param lp_index_3: Index of the LP on atom 3 used to make this hyperbond
        :hybrid_index_1: Index of the natural hybrid orbital (NHO) on atom 1 used to make this hyperbond
        :hybrid_index_2: Index of the natural hybrid orbital (NHO) on atom 2 used to make this hyperbond
        :hybrid_index_3: Index of the natural hybrid orbital (NHO) on atom 3 used to make this hyperbond
        """

        self.index = int(index)
        self.atom1_index = int(atom1_index)
        self.atom2_index = int(atom2_index)
        self.atom3_index = int(atom3_index)
        self.fraction_12 = float(fraction_12)
        self.fraction_23 = float(fraction_23)
        self.occupancy = float(occupancy)
        self.bond_index_12 = int(bond_index_12)
        self.lp_index_3 = int(lp_index_3)
        self.hybrid_index_1 = int(hybrid_index_1)
        self.hybrid_index_2 = int(hybrid_index_2)
        self.hybrid_index_3 = int(hybrid_index_3)


@set_msonable_type_adapter
class Interaction(MSONable):
    def __init__(
        self,
        perturbation_energy: float,
        energy_difference: float,
        fock_element: float,
        donor_index: int,
        acceptor_index: int,
        donor_type: str,
        acceptor_type: str,
        donor_atom1_index: int,
        acceptor_atom1_index: int,
        donor_atom2_index: int | None = None,
        acceptor_atom2_index: int | None = None,
        donor_atom3_index: int | None = None,
        acceptor_atom3_index: int | None = None,
    ):
        """
        Description of an interaction between two orbitals

        :param perturbation_energy: second-order perturbation energy, in kcal/mol
        :param energy_difference: difference in energy between the interacting orbitals, in Ha
        :param fock_element: Fock matrix element corresponding to this interaction, in a.u.
        :donor_index: Index of the donating orbital
        :acceptor_index: Index of the accepting orbital
        :donor_type: Type code for the donating orbital
        :acceptor_type: Type code for the accepting orbital
        :donor_atom1_index: Index of the first atom involved in the donor orbital
        :acceptor_atom1_index: Index of the first atom involved in the acceptor orbital
        :donor_atom2_index: Index of the second atom involved in the donor orbital
        :acceptor_atom2_index: Index of the second atom involved in the acceptor orbital
        :donor_atom3_index: Index of the third atom involved in the donor orbital
        :acceptor_atom3_index: Index of the third atom involved in the acceptor orbital
        """

        self.donor_index = int(donor_index)
        self.acceptor_index = int(acceptor_index)

        self.donor_type = donor_type
        self.acceptor_type = acceptor_type

        if isinstance(donor_atom2_index, int):
            if donor_atom2_index < 0:
                donor2 = None
            else:
                donor2 = int(donor_atom2_index)
        else:
            donor2 = None

        if isinstance(acceptor_atom2_index, int):
            if acceptor_atom2_index < 0:
                acceptor2 = None
            else:
                acceptor2 = int(acceptor_atom2_index)
        else:
            acceptor2 = None

        if isinstance(donor_atom3_index, int):
            # Not sure if this is actual possible
            if donor_atom3_index < 0:
                donor3 = None
            else:
                donor3 = int(donor_atom3_index)
        else:
            donor3 = None

        if isinstance(acceptor_atom3_index, int):
            # Similarly, not sure if this is important
            if acceptor_atom3_index < 0:
                acceptor3 = None
            else:
                acceptor3 = int(acceptor_atom3_index)
        else:
            acceptor3 = None

        self.donor_atom_indices = [
            index
            for index in [int(donor_atom1_index), donor2, donor3]
            if index is not None
        ]
        self.acceptor_atom_indices = [
            index
            for index in [int(acceptor_atom1_index), acceptor2, acceptor3]
            if index is not None
        ]

        self.perturbation_energy = float(perturbation_energy)
        self.energy_difference = float(energy_difference)
        self.fock_element = float(fock_element)

    def as_dict(self):
        return {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "donor_index": self.donor_index,
            "acceptor_index": self.acceptor_index,
            "donor_type": self.donor_type,
            "acceptor_type": self.acceptor_type,
            "donor_atom_indices": self.donor_atom_indices,
            "acceptor_atom_indices": self.acceptor_atom_indices,
            "perturbation_energy": self.perturbation_energy,
            "energy_difference": self.energy_difference,
            "fock_element": self.fock_element,
        }

    @classmethod
    def from_dict(cls, d):
        donor_inds = d["donor_atom_indices"]
        acceptor_inds = d["acceptor_atom_indices"]

        if len(donor_inds) < 2:
            donor_inds += [None, None]
        elif len(donor_inds) < 3:
            donor_inds += [None]

        if len(acceptor_inds) < 2:
            acceptor_inds += [None, None]
        elif len(acceptor_inds) < 3:
            acceptor_inds += [None]

        return cls(
            d["perturbation_energy"],
            d["energy_difference"],
            d["fock_element"],
            d["donor_index"],
            d["acceptor_index"],
            d["donor_type"],
            d["acceptor_type"],
            donor_inds[0],
            acceptor_inds[0],
            donor_inds[1],
            acceptor_inds[1],
            donor_inds[2],
            acceptor_inds[2],
        )


class OrbitalDoc(PropertyDoc):
    property_name: str = "natural bonding orbitals"

    # Always populated - closed-shell and open-shell
    open_shell: bool = Field(
        ..., description="Is this molecule open-shell (spin multiplicity != 1)?"
    )

    nbo_population: list[NaturalPopulation] = Field(
        ..., description="Natural electron populations of the molecule"
    )

    # Populated for closed-shell molecules
    nbo_lone_pairs: list[LonePair] | None = Field(
        None, description="Lone pair orbitals of a closed-shell molecule"
    )

    nbo_bonds: list[Bond] | None = Field(
        None, description="Bond-like orbitals of a closed-shell molecule"
    )

    nbo_three_center_bonds: list[ThreeCenterBond] | None = Field(
        None, description="Three-center bond-like orbitals of a closed-shell molecule"
    )

    nbo_hyperbonds: list[Hyperbond] | None = Field(
        None, description="3-center hyperbond-like orbitals of a closed-shell molecule"
    )

    nbo_interactions: list[Interaction] | None = Field(
        None, description="Orbital-orbital interactions of a closed-shell molecule"
    )

    # Populated for open-shell molecules
    alpha_population: list[NaturalPopulation] | None = Field(
        None,
        description="Natural electron populations of the alpha electrons of an "
        "open-shell molecule",
    )
    beta_population: list[NaturalPopulation] | None = Field(
        None,
        description="Natural electron populations of the beta electrons of an "
        "open-shell molecule",
    )

    alpha_lone_pairs: list[LonePair] | None = Field(
        None, description="Alpha electron lone pair orbitals of an open-shell molecule"
    )
    beta_lone_pairs: list[LonePair] | None = Field(
        None, description="Beta electron lone pair orbitals of an open-shell molecule"
    )

    alpha_bonds: list[Bond] | None = Field(
        None, description="Alpha electron bond-like orbitals of an open-shell molecule"
    )
    beta_bonds: list[Bond] | None = Field(
        None, description="Beta electron bond-like orbitals of an open-shell molecule"
    )

    alpha_three_center_bonds: list[ThreeCenterBond] | None = Field(
        None,
        description="Alpha electron three-center bond-like orbitals of an open-shell molecule",
    )
    beta_three_center_bonds: list[ThreeCenterBond] | None = Field(
        None,
        description="Beta electron three-center bond-like orbitals of an open-shell molecule",
    )

    alpha_hyperbonds: list[Hyperbond] | None = Field(
        None,
        description="Alpha electron hyperbond-like orbitals of an open-shell molecule",
    )
    beta_hyperbonds: list[Hyperbond] | None = Field(
        None,
        description="Beta electron hyperbond-like orbitals of an open-shell molecule",
    )

    alpha_interactions: list[Interaction] | None = Field(
        None,
        description="Alpha electron orbital-orbital interactions of an open-shell molecule",
    )
    beta_interactions: list[Interaction] | None = Field(
        None,
        description="Beta electron orbital-orbital interactions of an open-shell molecule",
    )

    @staticmethod
    def get_populations(nbo: dict[str, Any], indices: list[int]):
        """
        Helper function to extract natural population information
        from NBO output

        :param nbo: Dictionary of NBO output data
        :param indices: Data subsets from which to extract natural populations
        :return: population_sets (list of lists of NaturalPopulation)
        """

        population_sets = list()

        for pop_ind in indices:
            pops = nbo["natural_populations"][pop_ind]
            population = list()
            for ind, atom_num in pops["No"].items():
                population.append(
                    NaturalPopulation(
                        atom_num - 1,
                        pops["Core"][ind],
                        pops["Valence"][ind],
                        pops["Rydberg"][ind],
                        pops["Total"][ind],
                    )
                )
            population_sets.append(population)

        return population_sets

    @staticmethod
    def get_lone_pairs(nbo: dict[str, Any], indices: list[int]):
        """
        Helper function to extract lone pair information from NBO output

        :param nbo: Dictionary of NBO output data
        :param indices: Data subsets from which to extract lone pair information
        :return: lone_pairs (list of LonePairs)
        """

        lone_pair_sets = list()

        for lp_ind in indices:
            lps = nbo["hybridization_character"][lp_ind]
            lone_pairs = list()
            for ind, orb_ind in lps.get("bond index", dict()).items():
                this_lp = LonePair(
                    orb_ind,
                    lps["orbital index"][ind],
                    int(lps["atom number"][ind]) - 1,
                    lps["s"][ind],
                    lps["p"][ind],
                    lps["d"][ind],
                    lps["f"][ind],
                    lps["occupancy"][ind],
                    lps["type"][ind],
                )
                lone_pairs.append(this_lp)
            lone_pair_sets.append(lone_pairs)

        return lone_pair_sets

    @staticmethod
    def get_bonds(nbo: dict[str, Any], indices: list[int]):
        """
        Helper function to extract bonding information from NBO output

        :param nbo: Dictionary of NBO output data
        :param indices: Data subsets from which to extract bonds
        :return: bonds (list of Bonds)
        """

        bond_sets = list()

        for bd_ind in indices:
            bds = nbo["hybridization_character"][bd_ind]
            bonds = list()
            for ind, orb_ind in bds.get("bond index", dict()).items():
                this_bond = Bond(
                    orb_ind,
                    bds["orbital index"][ind],
                    int(bds["atom 1 number"][ind]) - 1,
                    int(bds["atom 2 number"][ind]) - 1,
                    bds["atom 1 s"][ind],
                    bds["atom 2 s"][ind],
                    bds["atom 1 p"][ind],
                    bds["atom 2 p"][ind],
                    bds["atom 1 d"][ind],
                    bds["atom 2 d"][ind],
                    bds["atom 1 f"][ind],
                    bds["atom 2 f"][ind],
                    bds["atom 1 polarization"][ind],
                    bds["atom 2 polarization"][ind],
                    bds["atom 1 pol coeff"][ind],
                    bds["atom 2 pol coeff"][ind],
                    bds["occupancy"][ind],
                    bds["type"][ind],
                )
                bonds.append(this_bond)
            bond_sets.append(bonds)

        return bond_sets

    @staticmethod
    def get_three_center_bonds(nbo: dict[str, Any], indices: list[int]):
        """
        Helper function to extract bonding information from NBO output

        :param nbo: Dictionary of NBO output data
        :param indices: Data subsets from which to extract bonds
        :return: threec_bonds (list of ThreeCenterBonds)
        """

        if len(indices) == 0:
            return None

        threec_sets = list()

        for tc_ind in indices:
            tcs = nbo["hybridization_character"][tc_ind]
            threecs = list()
            for ind, orb_ind in tcs.get("bond index", dict()).items():
                this_threec = ThreeCenterBond(
                    orb_ind,
                    tcs["orbital index"][ind],
                    int(tcs["atom 1 number"][ind]) - 1,
                    int(tcs["atom 2 number"][ind]) - 1,
                    int(tcs["atom 3 number"][ind]) - 1,
                    tcs["atom 1 s"][ind],
                    tcs["atom 2 s"][ind],
                    tcs["atom 3 s"][ind],
                    tcs["atom 1 p"][ind],
                    tcs["atom 2 p"][ind],
                    tcs["atom 3 p"][ind],
                    tcs["atom 1 d"][ind],
                    tcs["atom 2 d"][ind],
                    tcs["atom 3 d"][ind],
                    tcs["atom 1 f"][ind],
                    tcs["atom 2 f"][ind],
                    tcs["atom 3 f"][ind],
                    tcs["atom 1 polarization"][ind],
                    tcs["atom 2 polarization"][ind],
                    tcs["atom 3 polarization"][ind],
                    tcs["atom 1 pol coeff"][ind],
                    tcs["atom 2 pol coeff"][ind],
                    tcs["atom 3 pol coeff"][ind],
                    tcs["occupancy"][ind],
                    tcs["type"][ind],
                )
                threecs.append(this_threec)
            threec_sets.append(threecs)

        return threec_sets

    @staticmethod
    def get_hyperbonds(nbo: dict[str, Any], indices: list[int]):
        """
        Helper function to extract hyperbond information form NBO output

        :param nbo: Dictionary of NBO output data
        :param indices: Data subsets from which to extract interactions
        :return: interactions (list of Hyperbonds)
        """

        hyperbond_sets = list()

        # No hyperbonds present
        if "hyperbonds" not in nbo or len(indices) == 0 or len(nbo["hyperbonds"]) == 0:
            return None

        for hb_ind in indices:
            # For other types of bonds, all indices must be present
            # That is, we always expect one set of orbitals for closed-shell molecules,
            # And we always expect two sets (alpha and beta) for open-shell molecules
            # Hyperbonds are different.
            # Since these are (up to) 4-electron configurations, you could have a partially occupied
            # hyperbond with only two electrons
            # And in this case, for an open-shell molecule, you may only have one set of hyperbonds
            hyperbonds = list()
            try:
                hbds = nbo["hyperbonds"][hb_ind]
            except IndexError:
                hbds = dict()
            for ind, orb_ind in hbds.get("hyperbond index", dict()).items():
                this_hyperbond = Hyperbond(
                    orb_ind,
                    int(hbds["bond atom 1 index"][ind]) - 1,
                    int(hbds["bond atom 2 index"][ind]) - 1,
                    int(hbds["bond atom 3 index"][ind]) - 1,
                    hbds["pctA-B"][ind],
                    hbds["pctB-C"][ind],
                    hbds["occ"][ind],
                    hbds["BD(A-B)"][ind],
                    hbds["LP(C)"][ind],
                    hbds["h(A)"][ind],
                    hbds["h(B)"][ind],
                    hbds["h(C)"][ind],
                )
                hyperbonds.append(this_hyperbond)
            hyperbond_sets.append(hyperbonds)

        return hyperbond_sets

    @staticmethod
    def get_interactions(nbo: dict[str, Any], indices: list[int]):
        """
        Helper function to extract orbital interaction information
        from NBO output

        :param nbo: Dictionary of NBO output data
        :param indices: Data subsets from which to extract interactions
        :return: interactions (list of Interactions)
        """

        interaction_sets = list()

        for pert_ind in indices:
            perts = nbo["perturbation_energy"][pert_ind]
            interactions = list()
            for ind in perts.get("donor bond index", list()):
                donor_atom2 = perts["donor atom 2 number"].get(ind)
                if donor_atom2 == "info_is_from_3C":
                    # There's a pretty horrible hack in the current pymatgen NBO parsers
                    # To prevent a dramatic increase in storage space, atom indices and element symbols for 3C bonds
                    # are stored using the same keys as those used for lone pairs and conventional two-center bonds
                    donor_atom1_index = (
                        int(re.sub(r"\D", "", perts["donor atom 1 symbol"][ind])) - 1
                    )
                    donor_atom2_index = (
                        int(re.sub(r"\D", "", perts["donor atom 1 number"][ind])) - 1
                    )
                    donor_atom3_index = (
                        int(re.sub(r"\D", "", perts["donor atom 2 symbol"][ind])) - 1
                    )
                else:
                    donor_atom1_index = int(perts["donor atom 1 number"][ind]) - 1
                    donor_atom3_index = None
                    if donor_atom2 is None:
                        donor_atom2_index = None
                    else:
                        donor_atom2_index = int(donor_atom2) - 1

                acceptor_atom2 = perts["acceptor atom 2 number"].get(ind)
                if acceptor_atom2 == "info_is_from_3C":
                    acceptor_atom1_index = (
                        int(re.sub(r"\D", "", perts["acceptor atom 1 symbol"][ind])) - 1
                    )
                    acceptor_atom2_index = (
                        int(re.sub(r"\D", "", perts["acceptor atom 1 number"][ind])) - 1
                    )
                    acceptor_atom3_index = (
                        int(re.sub(r"\D", "", perts["acceptor atom 2 symbol"][ind])) - 1
                    )
                else:
                    acceptor_atom1_index = int(perts["acceptor atom 1 number"][ind]) - 1
                    acceptor_atom3_index = None
                    if acceptor_atom2 is None:
                        acceptor_atom2_index = None
                    else:
                        acceptor_atom2_index = int(acceptor_atom2) - 1

                this_inter = Interaction(
                    perts["perturbation energy"][ind],
                    perts["energy difference"][ind],
                    perts["fock matrix element"][ind],
                    int(perts["donor bond index"][ind]),
                    int(perts["acceptor bond index"][ind]),
                    perts["donor type"][ind],
                    perts["acceptor type"][ind],
                    donor_atom1_index,
                    acceptor_atom1_index,
                    donor_atom2_index,
                    acceptor_atom2_index,
                    donor_atom3_index,
                    acceptor_atom3_index,
                )

                interactions.append(this_inter)
            interaction_sets.append(interactions)

        return interaction_sets

    @classmethod
    def from_task(
        cls,
        task: TaskDocument,
        molecule_id: MPculeID,
        deprecated: bool = False,
        **kwargs,
    ):  # type: ignore[override]
        """
        Construct an orbital document from a task

        :param task: document from which vibrational properties can be extracted
        :param molecule_id: MPculeID
        :param deprecated: bool. Is this document deprecated?
        :param kwargs: to pass to PropertyDoc
        :return:
        """

        if task.output.nbo is None:
            raise ValueError("No NBO output in task {}!".format(task.task_id))
        elif not (
            task.orig["rem"].get("run_nbo6", False)
            or task.orig["rem"].get("nbo_external", False)
        ):
            raise ValueError("Only NBO7 is allowed!")

        nbo = task.output.nbo

        if task.output.optimized_molecule is not None:
            mol = task.output.optimized_molecule
        else:
            mol = task.output.initial_molecule

        spin = mol.spin_multiplicity

        # Closed-shell
        if int(spin) == 1:
            pops_inds = [0]
            lps_inds = [0]
            bds_inds = [1]
            if "hyperbonds" in nbo:
                # New parser, with three-center bonds
                tc_inds = [2]
            else:
                # Old parser - no three-center bonds
                tc_inds = list()
            hbds_inds = [0]
            perts_inds = [0]

        # Open-shell
        else:
            pops_inds = [0, 1, 2]
            if len(nbo.get("hybridization_character", [0, 0, 0, 0])) == 4:
                # Old parser - no three-center bonds
                lps_inds = [0, 2]
                bds_inds = [1, 3]
                tc_inds = list()
            else:
                # New parser - with three-center bonds
                lps_inds = [0, 3]
                bds_inds = [1, 4]
                tc_inds = [2, 5]
            hbds_inds = [0, 1]
            perts_inds = [0, 1]

        for dset, inds in [
            ("natural_populations", pops_inds),
            ("hybridization_character", lps_inds + bds_inds + tc_inds),
            ("perturbation_energy", perts_inds),
        ]:
            if len(nbo.get(dset, list())) <= inds[-1]:
                return

        population_sets = cls.get_populations(nbo, pops_inds)
        lone_pair_sets = cls.get_lone_pairs(nbo, lps_inds)
        bond_sets = cls.get_bonds(nbo, bds_inds)
        threec_sets = cls.get_three_center_bonds(nbo, tc_inds)
        hyperbond_sets = cls.get_hyperbonds(nbo, hbds_inds)
        interaction_sets = cls.get_interactions(nbo, perts_inds)

        if not (
            task.orig["rem"].get("run_nbo6")
            or task.orig["rem"].get("nbo_external", False)
        ):
            warnings = ["Using NBO5"]
        else:
            warnings = list()

        id_string = (
            f"natural_bonding_orbitals-{molecule_id}-{task.task_id}-{task.lot_solvent}"
        )
        h = blake2b()
        h.update(id_string.encode("utf-8"))
        property_id = h.hexdigest()

        if int(spin) == 1:
            if threec_sets is None:
                threec = None
            else:
                threec = threec_sets[0]

            if hyperbond_sets is None:
                hyperbond = None
            else:
                hyperbond = hyperbond_sets[0]

            return super().from_molecule(
                meta_molecule=mol,
                property_id=property_id,
                molecule_id=molecule_id,
                level_of_theory=task.level_of_theory,
                solvent=task.solvent,
                lot_solvent=task.lot_solvent,
                open_shell=False,
                nbo_population=population_sets[0],
                nbo_lone_pairs=lone_pair_sets[0],
                nbo_bonds=bond_sets[0],
                nbo_three_center_bonds=threec,
                nbo_hyperbonds=hyperbond,
                nbo_interactions=interaction_sets[0],
                origins=[
                    MolPropertyOrigin(
                        name="natural_bonding_orbitals", task_id=task.task_id
                    )
                ],
                warnings=warnings,
                deprecated=deprecated,
                **kwargs,
            )

        else:
            if threec_sets is None:
                threec_alpha = None
                threec_beta = None
            else:
                threec_alpha = threec_sets[0]
                threec_beta = threec_sets[1]

            if hyperbond_sets is None:
                hyperbond_alpha = None
                hyperbond_beta = None
            else:
                hyperbond_alpha = hyperbond_sets[0]
                hyperbond_beta = hyperbond_sets[1]

            return super().from_molecule(
                meta_molecule=mol,
                property_id=property_id,
                molecule_id=molecule_id,
                level_of_theory=task.level_of_theory,
                solvent=task.solvent,
                lot_solvent=task.lot_solvent,
                open_shell=True,
                nbo_population=population_sets[0],
                alpha_population=population_sets[1],
                beta_population=population_sets[2],
                alpha_lone_pairs=lone_pair_sets[0],
                beta_lone_pairs=lone_pair_sets[1],
                alpha_bonds=bond_sets[0],
                beta_bonds=bond_sets[1],
                alpha_three_center_bonds=threec_alpha,
                beta_three_center_bonds=threec_beta,
                alpha_hyperbonds=hyperbond_alpha,
                beta_hyperbonds=hyperbond_beta,
                alpha_interactions=interaction_sets[0],
                beta_interactions=interaction_sets[1],
                origins=[
                    MolPropertyOrigin(
                        name="natural bonding orbitals", task_id=task.task_id
                    )
                ],
                warnings=warnings,
                deprecated=deprecated,
                **kwargs,
            )



"""Test VASP parsing utilities."""

from tempfile import TemporaryDirectory
from pathlib import Path
import pytest

from emmet.core.vasp.utils import (
    recursive_discover_vasp_files,
    discover_and_sort_vasp_files,
    FileMetadata,
)


def test_file_meta(tmp_dir):
    import blake3

    incar_bytes = """
ALGO = Normal
ENCUT = 500
EDIFF = 0.0001
IBRION = -1
""".encode()

    with open(file_name := "INCAR.bz2", "wb") as f:
        f.write(incar_bytes)

    file_meta = FileMetadata(name="INCAR.bz2", path=file_name)
    file_meta.compute_hash()
    assert Path(file_meta.path).exists()
    assert file_meta.hash == blake3.blake3(incar_bytes).hexdigest()


def test_file_discovery():
    directory_structure = {
        "./neb_calc/00/": ["INCAR.gz", "KPOINTS", "POSCAR.gz"],
        "./neb_calc/01/": [
            "INCAR.gz",
            "CHGCAR",
            "CONTCAR.gz",
            "KPOINTS",
            "OUTCAR",
            "POSCAR.gz",
            "POTCAR.bz2",
            "vasprun.xml",
        ],
        "./neb_calc/02/": ["INCAR.gz", "KPOINTS", "POSCAR.gz"],
        "neb_calc": ["INCAR.gz", "KPOINTS", "POSCAR.gz", "POTCAR.lzma"],
        "block_2025_02_30/launcher_2025_02_31/launcher_2025_02_31_0001": [
            "AECCAR0.bz2",
            "LOCPOT.gz",
            "CONTCAR.relax1",
            "OUTCAR.relax1",
            "vasprun.xml.relax1",
            "POSCAR.T=300.gz",
            "POSCAR.T=1000.gz",
        ],
    }
    for idx in range(1, 3):
        directory_structure[
            "block_2025_02_30/launcher_2025_02_31/launcher_2025_02_31_0001"
        ].extend(
            [
                f"INCAR.relax{idx}",
                f"INCAR.orig.relax{idx}",
                f"POSCAR.relax{idx}",
                f"POSCAR.orig.relax{idx}",
                f"POTCAR.relax{idx}",
                f"POTCAR.orig.relax{idx}",
            ]
        )

    with TemporaryDirectory() as _tmp_dir:
        tmp_dir = Path(_tmp_dir).resolve()
        for calc_dir, files in directory_structure.items():
            p = tmp_dir / calc_dir
            if not p.exists():
                p.mkdir(parents=True, exist_ok=True)
            for f in files:
                (p / f).touch()
        vasp_files = recursive_discover_vasp_files(tmp_dir)
        valid_vasp_files = recursive_discover_vasp_files(tmp_dir, only_valid=True)

        assert len(recursive_discover_vasp_files(tmp_dir, max_depth=2)) == 4
        assert len(recursive_discover_vasp_files(tmp_dir, max_depth=1)) == 1

        files_by_calc_suffix = discover_and_sort_vasp_files(
            tmp_dir / "block_2025_02_30/launcher_2025_02_31/launcher_2025_02_31_0001"
        )

        for max_depth in (-1, 1.5):
            with pytest.raises(ValueError, match="non-negative integer"):
                _ = recursive_discover_vasp_files(tmp_dir, max_depth=max_depth)

    assert files_by_calc_suffix["relax1"]["contcar_file"].name == "CONTCAR.relax1"
    assert files_by_calc_suffix["relax1"]["vasprun_file"].name == "vasprun.xml.relax1"
    assert files_by_calc_suffix["relax1"]["outcar_file"].name == "OUTCAR.relax1"
    assert {b.name for b in files_by_calc_suffix["standard"]["volumetric_files"]} == {
        "AECCAR0.bz2",
        "LOCPOT.gz",
    }
    assert {b.name for b in files_by_calc_suffix["standard"]["elph_poscars"]} == {
        "POSCAR.T=300.gz",
        "POSCAR.T=1000.gz",
    }

    # should find all of the defined calculation directories + suffixes within those
    assert len(vasp_files) == 7

    found_files = set()
    ref_files = set()
    for calc_dir, files in directory_structure.items():
        ref_files.update({tmp_dir / calc_dir / file for file in files})
    for file_metas in vasp_files.values():
        found_files.update({file_meta.path for file_meta in file_metas})

    assert found_files == ref_files

    # Should only find two valid calculation directory / suffix pairs
    valid_calc_dirs = (
        "neb_calc/01/",
        "block_2025_02_30/launcher_2025_02_31/launcher_2025_02_31_0001",
    )

    assert len(valid_vasp_files) == 2
    found_valid_dirs = {calc_loc.path for calc_loc in valid_vasp_files}
    assert {tmp_dir / valid_dir for valid_dir in valid_calc_dirs} == found_valid_dirs



from importlib.resources import files as import_resource_file
from ruamel.yaml import YAML

with open(
    import_resource_file("emmet.core.vasp.calc_types") / "calc_types.yaml", "r"
) as f:
    config = YAML().load(f)

_REFERENCE_MEMBER_COUNT = {
    "RunType": 2 * sum(len(rtypes) for rtypes in config["RUN_TYPES"].values()),
    "TaskType": len(config["TASK_TYPES"]),
}
_REFERENCE_MEMBER_COUNT["CalcType"] = (
    _REFERENCE_MEMBER_COUNT["RunType"] * _REFERENCE_MEMBER_COUNT["TaskType"]
)


def test_enums():
    from emmet.core.vasp.calc_types import enums as vasp_enums

    for k, count in _REFERENCE_MEMBER_COUNT.items():
        # Test that we have expected number of enum members
        curr_enum = getattr(vasp_enums, k, None)
        assert len(curr_enum) == count

        if k == "RunType":
            # Test that RunType enum is not case sensitive
            assert all(
                curr_enum(member.value.upper()) == curr_enum(member.value.lower())
                for member in curr_enum
            )



"""Core definition of Spectrum document"""

from pydantic import Field

from emmet.core.base import EmmetBaseModel
from emmet.core.types.typing import DateTimeType, IdentifierType
from emmet.core.structure import StructureMetadata


class SpectrumDoc(StructureMetadata, EmmetBaseModel):
    """
    Base model definition for any spectra document. This should contain
    metadata on the structure the spectra pertains to
    """

    spectrum_name: str

    material_id: IdentifierType | None = Field(
        None,
        description="The ID of the material, used as a universal reference across proeprty documents. "
        "This comes in the form: mp-******.",
    )

    spectrum_id: str = Field(
        ...,
        title="Spectrum Document ID",
        description="The unique ID for this spectrum document.",
    )

    last_updated: DateTimeType = Field(
        description="Timestamp for the most recent calculation update for this property.",
    )

    warnings: list[str] = Field(
        [], description="Any warnings related to this property."
    )



from pymatgen.util.testing import STRUCTURES_DIR
from pymatgen.core import Structure

struct_names = (
    "SiO2 Li2O LiFePO4 TlBiSe2 K2O2 Li3V2(PO4)3 Li2O2 CsCl NaFePO4 Pb2TiZrO6 "
    "SrTiO3 TiO2 BaNiO3 VO2".split()
)

test_structures = {
    name: Structure.from_file(STRUCTURES_DIR / f"{name}.json") for name in struct_names
}



"""Define metrics of similarity between structures."""

from __future__ import annotations

from functools import partial
import multiprocessing
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from pydantic import BaseModel, Field

from emmet.core.material_property import PropertyDoc

try:
    from matminer.featurizers.structure.sites import SiteStatsFingerprint
    from matminer.featurizers.site.fingerprint import CrystalNNFingerprint
except ImportError:
    SiteStatsFingerprint = None
    CrystalNNFingerprint = None

try:
    import matgl
except ImportError:
    matgl = None  # type: ignore[assignment]


if TYPE_CHECKING:

    from matminer.featurizers.base import BaseFeaturizer
    from pymatgen.core import Structure


def _vector_difference_matrix_row(
    idxs: tuple[int, int],
    v: np.ndarray,
    norms: np.ndarray,
    dtype: np.dtype = np.dtype("float64"),
) -> tuple[tuple[int, int], np.ndarray]:
    """Compute the distance between a single vector and a list of other vectors.

    Parameters
    -----------
    idxs : the integer index of the vector in v to use as a fixed point
    v : the list of all possible vectors
    norms : the norms of the vectors in v
    dtype : the numpy dtype of working and returned arrays.

    Returns
    -----------
    int (copy of idxs) and np.ndarray of the squared vector distances.
    """
    inner = np.zeros((idxs[1] - idxs[0], v.shape[0]), dtype=dtype)
    _ = np.einsum("ik,jk->ij", v[idxs[0] : idxs[1]], v, out=inner)
    return idxs, np.array(
        [
            norms[idxs[0] : idxs[1]] + norms[i] - 2 * inner[:, i]
            for i in range(norms.shape[0])
        ],
        dtype=dtype,
    )


def vector_difference_matrix(
    v: np.ndarray,
    noise_floor: float = 1e-14,
    spread_rows: int = 0,
    dtype: np.dtype = np.dtype("float64"),
) -> np.ndarray:
    """Construct a symmetric matrix of vector differences.

    Given a list of vectors v, a symmetric matrix D such that:
        D_ij = | v_i - v_j |
        D_ji = D_ij

    if spread_rows is a positive int, then this will return only the
    upper triangle of D_ij, j >= i, and parallelize construction
    of the rows of D_ij.

    Parameters
    -----------
    v : numpy ndarray
        List of vectors. Axis = 0 should indicate distinct vectors,
        and axis = 1 their components.
    noise_floor : float = 1e-14
        Any values less than noise_floor will be zeroed out.
        Helps with loss of precision.
    spread_rows : int = 0
        The number of parallel processes to use in constructing
        the rows of D_ij
    dtype : the numpy dtype of the arrays used, defaults to float64
    """

    x = v.astype(dtype) if dtype != v.dtype else v.copy()
    vlen = x.shape[0]
    v_diff = np.zeros(2 * (vlen,), dtype=dtype)

    norms = np.zeros(vlen, dtype=dtype)
    _ = np.einsum("ik,ik->i", x, x, out=norms)

    if not spread_rows:
        inner = np.zeros(2 * (vlen,), dtype=dtype)
        _ = np.einsum("ik,jk->ij", x, x, out=inner)
        for i in range(vlen):
            v_diff[:, i] = norms + norms[i] - 2 * inner[:, i]
    else:
        func = partial(_vector_difference_matrix_row, v=x, norms=norms)
        with multiprocessing.Pool(spread_rows) as pool:
            vdiff_blocks = pool.map(
                func,
                [
                    (min(x), 1 + max(x))
                    for x in np.array_split(range(vlen), spread_rows)
                ],
            )

        for idxs, row in vdiff_blocks:
            v_diff[:, idxs[0] : idxs[1]] = row

    v_diff[v_diff < noise_floor] = 0.0
    return v_diff ** (0.5)


class SimilarityScorer:
    """Mixin for ranking the similarity between structures.

    Parameters
    -----------
    fingerprinter: BaseFeaturizer or None (default)
        A structural featurizer. If None, defaults to the
        featurizer used in the above reference.
    """

    def _featurize_structure(self, structure: Structure) -> np.ndarray:
        """Featurize a single structure using a user implemented method

        Parameters
        -----------
        structure : Structure

        Returns
        -----------
        np.ndarray
        """
        raise NotImplementedError

    def _post_process_distance(
        self,
        distances: np.ndarray,
    ) -> np.ndarray:
        """Postprocess vector distances to yield consistent similarity scores.

        This method should return a percentage where 0% indicates no
        similarity, and 100% indicates identical structures.

        Defaults to normalizing vector distances by tanh.
        See CrystalNNSimilarity for another example.

        Parameters
        -----------
        distances : a matrix of structure feature vector distances.

        Returns
        -----------
        np.ndarray
        """
        return 100 * (1.0 - np.tanh(distances))

    def featurize_structures(
        self,
        structures: list[Structure],
        num_procs: int = 1,
    ):
        """Featurize structures using the user-defined _featurize_structure.

        This method may be preferred when dealing with huge sets
        of structures. In those cases, getting distances between
        structures may lead to memory errors.

        Parameters
        -----------
        structures : list of Structure objects
        num_procs : int = 1
            Number of parallel processes to run in featurizing structures.

        Returns
        -----------
        np.ndarray : the feature vectors of the input structures.
        """
        if num_procs > 1:
            with multiprocessing.Pool(num_procs) as pool:
                _feature_vectors = pool.map(self._featurize_structure, structures)
        else:
            _feature_vectors = [
                self._featurize_structure(structure) for structure in structures
            ]

        return np.array(_feature_vectors)

    def _get_closest_vectors(
        self, idx: int, v: np.ndarray, num: int
    ) -> tuple[np.ndarray, np.ndarray]:
        """Return only a subset of vectors most similar to a specified vector.

        Parameters
        -----------
        idx : the specific index of the vector list to isolate
        v : numpy ndarray, the list of vectors
        num : the number of closest vectors to return

        """
        dist = np.linalg.norm(v[idx] - v, axis=1)
        idxs = np.array([j for j in np.argpartition(dist, num) if j != idx])[:num]

        subset_dist = dist[idxs]
        sorted_subset_idx = np.argsort(subset_dist)

        return idxs[sorted_subset_idx], self._post_process_distance(
            subset_dist[sorted_subset_idx]
        )

    def get_all_similarity_scores(
        self,
        structures: list[Structure],
        num_procs: int = 1,
        **kwargs,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Rank the similarity between structures using CrystalNN.

        Parameters
        -----------
        structures : list of Structure objects
        num_procs : int = 1
            Number of parallel processes to run in featurizing structures.
        **kwargs
            Kwargs to pass to `vector_difference_matrix`

        Returns
        -----------
        tuple of np.ndarray, np.ndarray
            The first array contains the structure feature vectors,
            and the second their similarity scores.
        """

        feature_vectors = self.featurize_structures(structures, num_procs=num_procs)
        distances = vector_difference_matrix(feature_vectors, **kwargs)
        return feature_vectors, self._post_process_distance(distances)

    def get_most_similar(
        self,
        feature_vectors: np.ndarray,
        num_procs: int = 1,
        num_top: int = 100,
        labels: list[str] | None = None,
    ) -> dict[str, dict[str, list[str] | np.ndarray]]:
        """Rank the similarity between structures using CrystalNN.

        Parameters
        -----------
        feature_vectors : list of feature vectors
        num_procs : int = 1
            Number of parallel processes to run in featurizing structures.
        num_top : int or None
            If an int, returns that number of most similar structures
            indicated by their indices in the original list.
            If None, returns all distances.
        labels : list of str or None
            If a list of str, the labels corresponding to the feature vectors,
            e.g., MPIDs.
            If None, defaults to the list indices.

        Returns
        -----------
        dict[int,dict[str,np.ndarray]]], containing the index of
            the structure in `structures`, with a dict containing
            the `indices` of the top `num_top` most similar
            structures and their corresponding `distances`.
        """
        wrapped = partial(self._get_closest_vectors, v=feature_vectors, num=num_top)
        nfv = feature_vectors.shape[0]
        with multiprocessing.Pool(num_procs) as pool:
            meta = pool.map(wrapped, range(nfv))
        labels = labels or [str(idx) for idx in range(nfv)]
        return {
            labels[idx]: {
                "indices": [labels[idx] for idx in field[0]],
                "distances": field[1],
            }
            for idx, field in enumerate(meta)
        }

    def build_similarity_collection_from_structures(
        self,
        structures: dict[str, Structure],
        num_procs: int = 1,
        num_top: int = 100,
    ) -> list[SimilarityDoc]:
        """Build a collection of similarity documents.

        This defines the build pipeline for the MP similarity collection.

        Parameters
        -----------
        structures : dict of str (e.g., MPID) to a corresponding structure.
        num_procs : int = 1
            Number of parallel processes to run in featurizing structures.
        num_top : int or None
            If an int, returns that number of most similar structures
            indicated by their indices in the original list.
            If None, returns all distances.

        Returns
        -----------
        A list of SimilarityDoc.
        """

        identifiers = list(structures)
        ordered_structures = [structures[idx] for idx in identifiers]
        feature_vectors = self.featurize_structures(
            ordered_structures, num_procs=num_procs
        )
        sim_meta = self.get_most_similar(
            feature_vectors, num_procs=num_procs, num_top=num_top, labels=identifiers
        )
        return [
            SimilarityDoc.from_structure(
                meta_structure=structures[idx],
                material_id=idx,
                feature_vector=feature_vectors[i],
                sim=[
                    SimilarityEntry(
                        task_id=jdx,
                        nelements=len(structures[jdx].composition.elements),
                        dissimilarity=100.0 - meta["distances"][j],  # type: ignore[operator]
                        formula=structures[jdx].formula,
                    )
                    for j, jdx in enumerate(meta["indices"])
                ],
            )
            for i, (idx, meta) in enumerate(sim_meta.items())
        ]

    @staticmethod
    def get_vendi_score(feature_vectors: np.ndarray) -> float:
        """Get the Vendi score of a set of feature vectors.

        Uses the conventions described in arXiv:2210.02410
        Describes the diveristy of a set of structures.

        Parameters
        -----------
        feature_vectors : np.ndarray
            The feature vectors, such as those from
            SimilarityScorer._featurize_structure
            Each row should be a distinct feature vector.

        Returns
        -----------
        float, the Vendi score.
            A Vendi score close to feature_vectors.shape[0] indicates
            high sample diversity, and a Vendi score close to
            1 indicates low sample diversity.
        """

        # Remove NaN rows
        non_nan_fv = feature_vectors[~np.any(np.isnan(feature_vectors), axis=1)]

        norms = np.einsum("ik,ik->i", non_nan_fv, non_nan_fv) ** (-0.5)
        norm_fv = np.einsum("ij,i->ij", non_nan_fv, norms)

        k_mat = np.einsum("ik,jk->ij", norm_fv, norm_fv)

        num_samp = non_nan_fv.shape[0]
        eigs = np.linalg.eigvalsh(k_mat) / num_samp
        log_eigs = np.zeros(num_samp)
        mask = eigs > 0.0
        log_eigs[mask] = np.log(eigs[mask])
        return np.exp(-np.einsum("i,i", eigs, log_eigs))


class CrystalNNSimilarity(SimilarityScorer):
    """Rank the similarity between structures using CrystalNN.

    Reference: 10.1039/C9RA07755C

    Parameters
    -----------
    fingerprinter: BaseFeaturizer or None (default)
        A structural featurizer. If None, defaults to the
        featurizer used in the above reference.
    """

    def __init__(self, fingerprinter: BaseFeaturizer | None = None) -> None:

        if fingerprinter is None and SiteStatsFingerprint is None:
            raise ImportError(
                "Please `pip install matminer` to use featurizer functionality."
            )

        self.fingerprinter = fingerprinter or SiteStatsFingerprint(
            CrystalNNFingerprint.from_preset(
                "ops",
                distance_cutoffs=None,
                x_diff_weight=None,
            ),
            stats=(
                "mean",
                "maximum",
            ),
        )
        self.num_feature = len(self.fingerprinter.feature_labels())

    def _featurize_structure(self, structure: Structure) -> np.ndarray:
        """Featurize a single structure using CrystalNN.

        Parameters
        -----------
        structure : Structure

        Returns
        -----------
        np.ndarray
        """

        try:
            return np.array(self.fingerprinter.featurize(structure))
        except Exception:
            return np.nan * np.ones(self.num_feature)

    def _post_process_distance(self, distances: np.ndarray) -> np.ndarray:
        """Use exponential weighting of feature vector distances.

        Parameters
        -----------
        distances : a matrix of structure feature vector distances.

        Returns
        -----------
        np.ndarray
        """
        return 100 * np.exp(-distances)


class M3GNetSimilarity(SimilarityScorer):
    """Obtain structural similarity using an M3GNet formation energy model.

    Parameters
    -----------
    model :  str or Path
        Model for computing feature vectors of a structure.
        Defaults to "M3GNet-MP-2018.6.1-Eform"
    """

    def __init__(self, model: str | Path = "M3GNet-MP-2018.6.1-Eform"):

        if matgl is None:
            raise ValueError("`pip install matgl` to use these features.")

        self.model = matgl.load_model(Path(model)).model

    def _featurize_structure(self, structure: Structure) -> np.ndarray:
        """Featurize a single structure using M3GNet-Eform.

        Parameters
        -----------
        structure : Structure

        Returns
        -----------
        np.ndarray
        """
        results = self.model.predict_structure(structure, return_features=True)
        return results["readout"].detach().cpu().numpy()[0]


class SimilarityEntry(BaseModel):
    """
    Find similar materials to a specified material based on crystal geometry.
    """

    task_id: str | None = Field(
        None,
        description="The Materials Project ID for the matched material. This comes in the form: mp-******.",
    )

    nelements: int | None = Field(
        None,
        description="Number of elements in the matched material.",
    )

    dissimilarity: float | None = Field(
        None,
        description="Dissimilarity measure for the matched material in %.",
    )

    formula: str | None = Field(
        None,
        description="Formula of the matched material.",
    )


class SimilarityDoc(PropertyDoc):
    """
    Model for a document containing structure similarity data
    """

    property_name: str = "similarity"

    sim: list[SimilarityEntry] | None = Field(
        None,
        description="List containing similar structure data for a given material.",
    )

    material_id: str | None = Field(
        None,
        description="The Materials Project ID for the material. This comes in the form: mp-******",
    )

    feature_vector: list[float] | None = Field(
        None, description="The feature / embedding vector of the structure."
    )



from typing import Annotated, TypeVar

from pydantic import BeforeValidator, WrapSerializer
from pymatgen.core.trajectory import Trajectory
from typing_extensions import TypedDict

from emmet.core.math import Matrix3D, Vector3D
from emmet.core.types.pymatgen_types.properties import TypedAggregateProperitesDict


class FrameProps(TypedDict):
    energy: float


TypedTrajectoryDict = TypedDict(
    "TypedTrajectoryDict",
    {
        "@module": str,
        "@class": str,
        "species": list[str],
        "coords": list[list[Vector3D]],
        "charge": float,
        "spin_multiplicity": float,
        "lattice": list[Matrix3D],
        "site_properties": TypedAggregateProperitesDict,
        "frame_properties": FrameProps,
        "constant_lattice": bool,
        "time_step": float,
        "coords_are_displacement": bool,
        "base_positions": list[Vector3D],
    },
)

TrajectoryTypeVar = TypeVar("TrajectoryTypeVar", Trajectory, TypedTrajectoryDict)

TrajectoryType = Annotated[
    TrajectoryTypeVar,
    BeforeValidator(lambda x: Trajectory.from_dict(x) if isinstance(x, dict) else x),
    WrapSerializer(lambda x, nxt, info: x.as_dict(), return_type=TypedTrajectoryDict),
]



"""Core definition of an Electronic Structure"""

from __future__ import annotations

from collections import defaultdict
from math import isnan
from typing import TYPE_CHECKING, Annotated, Literal, TypeVar

import numpy as np
from pydantic import BaseModel, BeforeValidator, Field, WrapSerializer, field_validator
from pymatgen.analysis.magnetism.analyzer import (
    CollinearMagneticStructureAnalyzer,
    Ordering,
)
from pymatgen.core import Structure
from pymatgen.electronic_structure.bandstructure import BandStructureSymmLine
from pymatgen.electronic_structure.core import OrbitalType, Spin
from pymatgen.electronic_structure.dos import CompleteDos
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.symmetry.bandstructure import HighSymmKpath

from emmet.core.material_property import PropertyDoc
from emmet.core.mpid import AlphaID
from emmet.core.settings import EmmetSettings
from emmet.core.types.enums import ValueEnum
from emmet.core.types.pymatgen_types.bandstructure_symm_line_adapter import (
    BandStructureSymmLineType,
)
from emmet.core.types.pymatgen_types.dos_adapter import CompleteDosType
from emmet.core.types.pymatgen_types.element_adapter import ElementType
from emmet.core.types.typing import DateTimeType, IdentifierType, TypedBandDict

if TYPE_CHECKING:
    from typing_extensions import Self

SETTINGS = EmmetSettings()

OrderingType = Annotated[
    Ordering,
    BeforeValidator(lambda x: Ordering(x) if isinstance(x, str) else x),
    WrapSerializer(lambda x, nxt, info: x.value, return_type=str),
]


class BSPathType(ValueEnum):
    setyawan_curtarolo = "setyawan_curtarolo"
    hinuma = "hinuma"
    latimer_munro = "latimer_munro"


class DOSProjectionType(ValueEnum):
    total = "total"
    elemental = "elemental"
    orbital = "orbital"


class BSObjectDoc(BaseModel):
    """
    Band object document.
    """

    task_id: IdentifierType | None = Field(
        None,
        description="The source calculation (task) ID that this band structure comes from. "
        "This has the same form as a Materials Project ID.",
    )

    last_updated: DateTimeType = Field(
        description="The timestamp when this calculation was last updated",
    )

    data: BandStructureSymmLineType | None = Field(
        None, description="The band structure object for the given calculation ID"
    )


class DOSObjectDoc(BaseModel):
    """
    DOS object document.
    """

    task_id: IdentifierType | None = Field(
        None,
        description="The source calculation (task) ID that this density of states comes from. "
        "This has the same form as a Materials Project ID.",
    )

    last_updated: DateTimeType = Field(
        description="The timestamp when this calculation was last updated.",
    )

    data: CompleteDosType | None = Field(
        None, description="The density of states object for the given calculation ID."
    )


class ElectronicStructureBaseData(BaseModel):
    task_id: IdentifierType = Field(
        ...,
        description="The source calculation (task) ID for the electronic structure data. "
        "This has the same form as a Materials Project ID.",
    )

    band_gap: float = Field(..., description="Band gap energy in eV.")
    cbm: float | None = Field(None, description="Conduction band minimum data.")
    vbm: float | None = Field(None, description="Valence band maximum data.")
    efermi: float | None = Field(None, description="Fermi energy in eV.")


class ElectronicStructureSummary(ElectronicStructureBaseData):
    is_gap_direct: bool = Field(..., description="Whether the band gap is direct.")
    is_metal: bool = Field(..., description="Whether the material is a metal.")
    magnetic_ordering: OrderingType = Field(
        ..., description="Magnetic ordering of the calculation."
    )


class BandStructureSummaryData(ElectronicStructureSummary):
    equivalent_labels: dict[str, dict[str, dict[str, str]]] = Field(
        ..., description="Equivalent k-point labels in other k-path conventions."
    )

    nbands: float = Field(..., description="Number of bands.")
    direct_gap: float = Field(..., description="Direct gap energy in eV.")
    cbm: TypedBandDict | None = Field(None, description="Conduction band minimum data.")
    vbm: TypedBandDict | None = Field(None, description="Valence band maximum data.")

    @field_validator("cbm", "vbm", mode="before")
    def deserialize_bands(cls, band):
        if band:
            if isinstance(band["kpoint"], dict) and "label" not in band["kpoint"]:
                band["kpoint"]["label"] = None
        return band

    @field_validator("equivalent_labels", mode="before")
    def deserialize_labels(cls, equivalent_labels):
        if isinstance(next(iter(equivalent_labels.values())), list):
            equivalent_labels = {
                convention: {
                    other_convention: {k: v for k, v in label_tuples}
                    for other_convention, label_tuples in other_mapping
                }
                for convention, other_mapping in equivalent_labels.items()
            }
        return equivalent_labels


class DosSummaryData(ElectronicStructureBaseData):
    spin_polarization: float | None = Field(
        None, description="Spin polarization at the fermi level."
    )


class BandstructureData(BaseModel):
    setyawan_curtarolo: BandStructureSummaryData | None = Field(
        None,
        description="Band structure summary data using the Setyawan-Curtarolo path convention.",
    )

    hinuma: BandStructureSummaryData | None = Field(
        None,
        description="Band structure summary data using the Hinuma et al. path convention.",
    )

    latimer_munro: BandStructureSummaryData | None = Field(
        None,
        description="Band structure summary data using the Latimer-Munro path convention.",
    )


SpinTypeVar = TypeVar("SpinTypeVar", Spin, Literal["1", "-1"])

SpinType = Annotated[
    SpinTypeVar,
    BeforeValidator(lambda x: Spin(int(x)) if isinstance(x, str) else x),
    WrapSerializer(lambda x, nxt, info: str(x), return_type=str),
]


class DosData(BaseModel):
    total: dict[SpinType, DosSummaryData] | None = Field(
        None, description="Total DOS summary data."
    )

    elemental: (
        dict[
            ElementType,
            dict[Literal["s", "p", "d", "f", "total"], dict[SpinType, DosSummaryData]],
        ]
        | None
    ) = Field(
        None,
        description="Band structure summary data using the Hinuma et al. path convention.",
    )

    orbital: (
        dict[Literal["s", "p", "d", "f", "total"], dict[SpinType, DosSummaryData]]
        | None
    ) = Field(
        None,
        description="Band structure summary data using the Latimer-Munro path convention.",
    )

    magnetic_ordering: OrderingType | None = Field(
        None, description="Magnetic ordering of the calculation."
    )

    @field_validator("elemental", mode="before")
    def elemental_deserializer(cls, elemental):
        if isinstance(next(iter(elemental.values())), list):
            elemental = {
                element: {
                    oribital_type: {
                        spin: summary_data for spin, summary_data in spin_summary_pairs
                    }
                    for oribital_type, spin_summary_pairs in orbital
                }
                for element, orbital in elemental.items()
            }

        return elemental

    @field_validator("orbital", mode="before")
    def oribital_deserializer(cls, orbital):
        if isinstance(next(iter(orbital.values())), list):
            orbital = {
                oribital_type: {
                    spin: summary_data for spin, summary_data in spin_summary_pairs
                }
                for oribital_type, spin_summary_pairs in orbital.items()
            }

        return orbital


class ElectronicStructureDoc(PropertyDoc, ElectronicStructureSummary):
    """
    Definition for a core Electronic Structure Document
    """

    property_name: str = "electronic_structure"

    bandstructure: BandstructureData | None = Field(
        None, description="Band structure data for the material."
    )

    dos: DosData | None = Field(
        None, description="Density of states data for the material."
    )

    @classmethod
    def from_bsdos(  # type: ignore[override]
        cls,
        dos: dict[IdentifierType, CompleteDos],
        is_gap_direct: bool,
        is_metal: bool,
        material_id: IdentifierType | None = None,
        origins: list[dict] = [],
        structures: dict[IdentifierType, Structure] | None = None,
        setyawan_curtarolo: dict[IdentifierType, BandStructureSymmLine] | None = None,
        hinuma: dict[IdentifierType, BandStructureSymmLine] | None = None,
        latimer_munro: dict[IdentifierType, BandStructureSymmLine] | None = None,
        **kwargs,
    ) -> Self:
        """
        Builds a electronic structure document using band structure and density of states data.

        Args:
            material_id (AlphaID or MPID): A material ID.
            dos (dict[AlphaID or MPID, CompleteDos]): Dictionary mapping a calculation (task) ID to a CompleteDos object.
            is_gap_direct (bool): Direct gap indicator included at root level of document.
            is_metal (bool): Metallic indicator included at root level of document.
            structures (dict[AlphaID or MPID, Structure]) = Dictionary mapping a calculation (task) ID to the structures used
                as inputs. This is to ensures correct magnetic moment information is included.
            setyawan_curtarolo (dict[AlphaID or MPID, BandStructureSymmLine]): Dictionary mapping a calculation (task) ID to a
                BandStructureSymmLine object from a calculation run using the Setyawan-Curtarolo k-path convention.
            hinuma (dict[AlphaID or MPID, BandStructureSymmLine]): Dictionary mapping a calculation (task) ID to a
                BandStructureSymmLine object from a calculation run using the Hinuma et al. k-path convention.
            latimer_munro (dict[AlphaID or MPID, BandStructureSymmLine]): Dictionary mapping a calculation (task) ID to a
                BandStructureSymmLine object from a calculation run using the Latimer-Munro k-path convention.
            origins (list[dict]): Optional origins information for final doc

        """

        # -- Process density of states data

        dos_task, dos_obj = list(dos.items())[0]

        orbitals = [OrbitalType.s, OrbitalType.p, OrbitalType.d]
        spins = list(dos_obj.densities.keys())

        ele_dos = dos_obj.get_element_dos()
        tot_orb_dos = dos_obj.get_spd_dos()

        elements = ele_dos.keys()

        dos_efermi = dos_obj.efermi

        is_gap_direct = is_gap_direct
        is_metal = is_metal

        structure = dos_obj.structure

        if structures is not None and structures[dos_task]:
            structure = structures[dos_task]

        dos_mag_ordering = CollinearMagneticStructureAnalyzer(structure).ordering

        dos_data = {
            "total": defaultdict(dict),
            "elemental": {element: defaultdict(dict) for element in elements},
            "orbital": defaultdict(dict),
            "magnetic_ordering": dos_mag_ordering,
        }

        for spin in spins:
            # - Process total DOS data
            band_gap = dos_obj.get_gap(spin=spin)
            (cbm, vbm) = dos_obj.get_cbm_vbm(spin=spin)

            try:
                spin_polarization = dos_obj.spin_polarization
                if spin_polarization is None or isnan(spin_polarization):
                    spin_polarization = None
            except KeyError:
                spin_polarization = None

            dos_data["total"][spin] = DosSummaryData(  # type: ignore[index]
                task_id=dos_task,
                band_gap=band_gap,
                cbm=cbm,
                vbm=vbm,
                efermi=dos_efermi,
                spin_polarization=spin_polarization,
            )

            # - Process total orbital projection data
            for orbital in orbitals:
                band_gap = tot_orb_dos[orbital].get_gap(spin=spin)

                (cbm, vbm) = tot_orb_dos[orbital].get_cbm_vbm(spin=spin)

                spin_polarization = None

                dos_data["orbital"][str(orbital)][spin] = DosSummaryData(  # type: ignore[index]
                    task_id=dos_task,
                    band_gap=band_gap,
                    cbm=cbm,
                    vbm=vbm,
                    efermi=dos_efermi,
                    spin_polarization=spin_polarization,
                )

        # - Process element and element orbital projection data
        for ele in ele_dos:
            orb_dos = dos_obj.get_element_spd_dos(ele)

            for orbital in ["total"] + list(orb_dos.keys()):  # type: ignore[assignment]
                if orbital == "total":
                    proj_dos = ele_dos
                    label = ele
                else:
                    proj_dos = orb_dos
                    label = orbital

                for spin in spins:
                    band_gap = proj_dos[label].get_gap(spin=spin)
                    (cbm, vbm) = proj_dos[label].get_cbm_vbm(spin=spin)

                    spin_polarization = None

                    dos_data["elemental"][ele][str(orbital)][spin] = DosSummaryData(  # type: ignore[index]
                        task_id=dos_task,
                        band_gap=band_gap,
                        cbm=cbm,
                        vbm=vbm,
                        efermi=dos_efermi,
                        spin_polarization=spin_polarization,
                    )

        #  -- Process band structure data
        bs_data = {  # type: ignore
            "setyawan_curtarolo": setyawan_curtarolo,
            "hinuma": hinuma,
            "latimer_munro": latimer_munro,
        }

        for bs_type, bs_input in bs_data.items():
            if bs_input is not None:
                bs_task, bs = list(bs_input.items())[0]

                if structures is not None and structures[bs_task]:
                    bs_mag_ordering = CollinearMagneticStructureAnalyzer(
                        structures[bs_task]
                    ).ordering
                else:
                    bs_mag_ordering = CollinearMagneticStructureAnalyzer(
                        bs.structure  # type: ignore[arg-type]
                    ).ordering

                gap_dict = bs.get_band_gap()
                is_metal = bs.is_metal()
                direct_gap = bs.get_direct_band_gap()

                if is_metal:
                    band_gap = 0.0
                    cbm = None  # type: ignore[assignment]
                    vbm = None  # type: ignore[assignment]
                    is_gap_direct = False
                else:
                    band_gap = gap_dict["energy"]
                    cbm = bs.get_cbm()  # type: ignore[assignment]
                    vbm = bs.get_vbm()  # type: ignore[assignment]
                    is_gap_direct = gap_dict["direct"]

                    # coerce type here, mixture of str and int types in bs objects
                    cbm["kpoint_index"] = [int(x) for x in cbm["kpoint_index"]]  # type: ignore[index]
                    vbm["kpoint_index"] = [int(x) for x in vbm["kpoint_index"]]  # type: ignore[index]

                bs_efermi = bs.efermi
                nbands = bs.nb_bands

                # - Get equivalent labels between different conventions
                hskp = HighSymmKpath(
                    bs.structure,
                    path_type="all",
                    symprec=0.1,
                    angle_tolerance=5,
                    atol=1e-5,
                )
                equivalent_labels = hskp.equiv_labels

                if bs_type == "latimer_munro":
                    gen_labels = set(
                        [
                            label
                            for label in equivalent_labels["latimer_munro"][
                                "setyawan_curtarolo"
                            ]
                        ]
                    )
                    kpath_labels = set(
                        [
                            kpoint.label
                            for kpoint in bs.kpoints
                            if kpoint.label is not None
                        ]
                    )

                    if not gen_labels.issubset(kpath_labels):
                        new_structure = SpacegroupAnalyzer(
                            bs.structure  # type: ignore[arg-type]
                        ).get_primitive_standard_structure(
                            international_monoclinic=False
                        )

                        hskp = HighSymmKpath(
                            new_structure,
                            path_type="all",
                            symprec=SETTINGS.SYMPREC,
                            angle_tolerance=SETTINGS.ANGLE_TOL,
                            atol=1e-5,
                        )
                        equivalent_labels = hskp.equiv_labels

                bs_data[bs_type] = BandStructureSummaryData(  # type: ignore
                    task_id=bs_task,
                    band_gap=band_gap,
                    direct_gap=direct_gap,
                    cbm=cbm,
                    vbm=vbm,
                    is_gap_direct=is_gap_direct,
                    is_metal=is_metal,
                    efermi=bs_efermi,
                    nbands=nbands,
                    equivalent_labels=equivalent_labels,
                    magnetic_ordering=bs_mag_ordering,
                )

        bs_entry = BandstructureData(**bs_data)  # type: ignore
        dos_entry = DosData(**dos_data)  # type: ignore[arg-type]

        # Obtain summary data

        bs_gap = (
            bs_entry.setyawan_curtarolo.band_gap
            if bs_entry.setyawan_curtarolo is not None
            else None
        )
        dos_cbm, dos_vbm = dos_obj.get_cbm_vbm()
        dos_gap = max(dos_cbm - dos_vbm, 0.0)

        new_origin_last_updated = None
        new_origin_task_id = None

        if bs_gap is not None and bs_gap <= dos_gap + 0.2:
            summary_task = bs_entry.setyawan_curtarolo.task_id  # type: ignore
            summary_band_gap = bs_gap
            summary_cbm = (
                bs_entry.setyawan_curtarolo.cbm.get("energy", None)  # type: ignore
                if bs_entry.setyawan_curtarolo.cbm is not None  # type: ignore
                else None
            )
            summary_vbm = (
                bs_entry.setyawan_curtarolo.vbm.get("energy", None)  # type: ignore
                if bs_entry.setyawan_curtarolo.cbm is not None  # type: ignore
                else None
            )  # type: ignore
            summary_efermi = bs_entry.setyawan_curtarolo.efermi  # type: ignore
            is_gap_direct = bs_entry.setyawan_curtarolo.is_gap_direct  # type: ignore
            is_metal = bs_entry.setyawan_curtarolo.is_metal  # type: ignore

            for origin in origins:
                if origin["name"] == "setyawan_curtarolo":
                    new_origin_last_updated = origin["last_updated"]
                    new_origin_task_id = origin["task_id"]

        else:
            summary_task = dos_entry.model_dump()["total"][str(Spin.up)]["task_id"]
            summary_band_gap = dos_gap
            summary_cbm = dos_cbm
            summary_vbm = dos_vbm
            summary_efermi = dos_efermi
            is_metal = True if np.isclose(dos_gap, 0.0, atol=0.01, rtol=0) else False

            for origin in origins:
                if origin["name"] == "dos":
                    new_origin_last_updated = origin["last_updated"]
                    new_origin_task_id = origin["task_id"]

        if new_origin_task_id is not None:
            for origin in origins:
                if origin["name"] == "electronic_structure":
                    origin["last_updated"] = new_origin_last_updated
                    origin["task_id"] = new_origin_task_id

        summary_magnetic_ordering = CollinearMagneticStructureAnalyzer(
            kwargs["meta_structure"],
            round_magmoms=True,
            threshold_nonmag=0.2,
            threshold=0,
        ).ordering

        return cls.from_structure(
            material_id=AlphaID(material_id),
            task_id=summary_task,
            band_gap=summary_band_gap,
            cbm=summary_cbm,
            vbm=summary_vbm,
            efermi=summary_efermi,
            is_gap_direct=is_gap_direct,
            is_metal=is_metal,
            magnetic_ordering=summary_magnetic_ordering,
            bandstructure=bs_entry,
            dos=dos_entry,
            **kwargs,
        )



import pytest

from emmet.core.testing_utils import assert_schemas_equal

from tests.conftest_qchem import get_test_object, OPENBABEL_INSTALLED


@pytest.mark.parametrize(
    "object_name, task_name",
    [
        pytest.param("SinglePointTest", "standard", id="SinglePointTest"),
        pytest.param("OptimizationTest", "standard", id="OptimizationTest"),
    ],
)  # Can add more later, something like freq, pesscan, ts,
# FFOptTest, once we get flows working for qchem in atomate2


def test_input_summary(test_dir, object_name, task_name):
    from monty.json import MontyDecoder, jsanitize
    from emmet.core.qc_tasks import InputDoc
    from emmet.core.qchem.calculation import Calculation

    test_object = get_test_object(object_name)
    dir_name = test_dir / "qchem" / test_object.folder

    files = test_object.task_files[task_name]
    calc_doc = Calculation.from_qchem_files(dir_name, task_name, **files)

    test_doc = InputDoc.from_qchem_calc_doc(calc_doc)
    valid_doc = test_object.task_doc["input"]
    assert_schemas_equal(test_doc, valid_doc)

    # test document can be jsanitized
    d = jsanitize(test_doc, strict=True, enum_values=True, allow_bson=True)

    # and decoded
    MontyDecoder().process_decoded(d)


@pytest.mark.parametrize(
    "object_name, task_name",
    [
        pytest.param("SinglePointTest", "standard", id="SinglePointTest"),
        pytest.param("OptimizationTest", "standard", id="OptimizationTest"),
    ],
)
def test_output_summary(test_dir, object_name, task_name):
    from monty.json import MontyDecoder, jsanitize
    from emmet.core.qc_tasks import OutputDoc
    from emmet.core.qchem.calculation import Calculation

    test_object = get_test_object(object_name)
    dir_name = test_dir / "qchem" / test_object.folder

    files = test_object.task_files[task_name]
    calc_doc = Calculation.from_qchem_files(dir_name, task_name, **files)

    test_doc = OutputDoc.from_qchem_calc_doc(calc_doc)
    valid_doc = test_object.task_doc["output"]
    assert_schemas_equal(test_doc, valid_doc)

    # test document can be janitized
    d = jsanitize(test_doc, strict=True, enum_values=True, allow_bson=True)

    # and decoded
    MontyDecoder().process_decoded(d)


@pytest.mark.parametrize(
    "object_name",
    [
        pytest.param("SinglePointTest", id="SinglePointTest"),
        pytest.param("OptimizationTest", id="OptimizationTest"),
    ],
)
def test_task_doc(test_dir, object_name):
    from monty.json import MontyDecoder, jsanitize
    from emmet.core.qc_tasks import TaskDoc

    test_object = get_test_object(object_name)
    dir_name = test_dir / "qchem" / test_object.folder
    test_doc = TaskDoc.from_directory(dir_name)
    assert_schemas_equal(test_doc, test_object.task_doc)

    # test document can be jsanitized
    d = jsanitize(test_doc, strict=True, enum_values=True, allow_bson=True)

    # and decoded
    if not OPENBABEL_INSTALLED:
        d.pop("custodian", None)
    MontyDecoder().process_decoded(d)

    # Test that additional_fields works
    test_doc = TaskDoc.from_directory(dir_name, additional_fields={"foo": "bar"})
    assert test_doc.model_dump()["additional_fields"] == {"foo": "bar"}


@pytest.mark.parametrize(
    "object_name",
    [
        pytest.param("SinglePointTest", id="SinglePointTest"),
        pytest.param("OptimizationTest", id="OptimizationTest"),
    ],
)
def test_task_doc_val_flag(test_dir, object_name):
    from monty.json import MontyDecoder, jsanitize
    from emmet.core.qc_tasks import TaskDoc

    test_object = get_test_object(object_name)
    dir_name = test_dir / "qchem" / test_object.folder
    test_doc = TaskDoc.from_directory(dir_name, validate_lot=False)
    assert_schemas_equal(test_doc, test_object.task_doc)

    # test document can be jsanitized
    d = jsanitize(test_doc, strict=True, enum_values=True, allow_bson=True)

    # and decoded
    if not OPENBABEL_INSTALLED:
        d.pop("custodian", None)
    MontyDecoder().process_decoded(d)

    # Test that additional_fields works
    test_doc = TaskDoc.from_directory(
        dir_name, validate_lot=False, additional_fields={"foo": "bar"}
    )
    assert test_doc.model_dump()["additional_fields"] == {"foo": "bar"}



import json
import datetime

import pytest

from monty.io import zopen

from emmet.core.qchem.task import TaskDocument
from emmet.core.molecules.electric import ElectricMultipoleDoc


__author__ = "Evan Spotte-Smith <ewcspottesmith@lbl.gov>"


@pytest.fixture(scope="session")
def test_tasks(test_dir):
    with zopen(test_dir / "multipole_docs.json.gz") as f:
        data = json.load(f)

    for d in data:
        d["last_updated"] = datetime.datetime.strptime(
            d["last_updated"]["string"], "%Y-%m-%d %H:%M:%S.%f"
        )

    tasks = [TaskDocument(**t) for t in data]
    return tasks


def test_electric_multipole(test_tasks):
    # First, test from SP
    sp_doc = ElectricMultipoleDoc.from_task(
        test_tasks[0], molecule_id="b9ba54febc77d2a9177accf4605767db-C1Li2O3-1-2"
    )
    assert sp_doc.property_name == "multipole_moments"
    assert sp_doc.total_dipole == pytest.approx(4.2371)
    assert sp_doc.dipole_moment[0] == pytest.approx(-3.2443)
    assert sp_doc.dipole_moment[1] == pytest.approx(-2.4294)
    assert sp_doc.dipole_moment[2] == pytest.approx(-1.2352)
    assert sp_doc.resp_total_dipole == pytest.approx(4.2845)
    assert len(sp_doc.quadrupole_moment) == 6
    assert len(sp_doc.octopole_moment) == 10
    assert len(sp_doc.hexadecapole_moment) == 15
    assert sp_doc.quadrupole_moment["XX"] == pytest.approx(-55.6651)
    assert sp_doc.octopole_moment["XXZ"] == pytest.approx(9.7933)
    assert sp_doc.hexadecapole_moment["YYZZ"] == pytest.approx(-78.2281)

    # Test from force calc
    force_doc = ElectricMultipoleDoc.from_task(
        test_tasks[1], molecule_id="b9ba54febc77d2a9177accf4605767db-C1Li2O3-1-2"
    )
    assert force_doc.total_dipole == pytest.approx(6.562)
    assert force_doc.dipole_moment[0] == pytest.approx(5.4663)

    # Test from FFOpt
    ffopt_doc = ElectricMultipoleDoc.from_task(
        test_tasks[2], molecule_id="b9ba54febc77d2a9177accf4605767db-C1Li2O3-1-2"
    )
    assert ffopt_doc.quadrupole_moment["XZ"] == pytest.approx(0.1372)

    # Test from opt
    opt_doc = ElectricMultipoleDoc.from_task(
        test_tasks[3], molecule_id="b9ba54febc77d2a9177accf4605767db-C1Li2O3-1-2"
    )
    assert opt_doc.octopole_moment["YZZ"] == pytest.approx(35.289)
    assert opt_doc.total_dipole == ffopt_doc.total_dipole

    # Test from freq
    freq_doc = ElectricMultipoleDoc.from_task(
        test_tasks[4], molecule_id="b9ba54febc77d2a9177accf4605767db-C1Li2O3-1-2"
    )
    assert freq_doc.hexadecapole_moment["ZZZZ"] == pytest.approx(-366.0089)



"""Define custom type annotations for emmet-core.

Note that only type annotations which are used across
the code base should be put here.

Types which only have one purpose / exist only within
one module, can and should remain in that module.
"""

from __future__ import annotations

import os
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Union

from pydantic import BeforeValidator, Field, PlainSerializer
from typing_extensions import TypedDict

from emmet.core.mpid import MPID, AlphaID
from emmet.core.types.pymatgen_types.kpoint_adapter import KpointType
from emmet.core.utils import convert_datetime, utcnow

if TYPE_CHECKING:
    from typing import Any

    from typing_extensions import TypeAlias

FSPathType: TypeAlias = Annotated[
    Union[str | Path | os.DirEntry[str] | os.PathLike[str]],
    PlainSerializer(lambda x: str(x), return_type=str),
]
"""Type of a generic path-like object"""


DateTimeType: TypeAlias = Annotated[
    datetime,
    Field(default_factory=utcnow),
    BeforeValidator(convert_datetime),
]
"""Datetime serde."""

NullableDateTimeType: TypeAlias = Annotated[
    datetime | None,
    Field(default_factory=utcnow),
    BeforeValidator(convert_datetime),
]
"""Nullable datetime serde.

See: https://docs.pydantic.dev/latest/concepts/fields/#the-annotated-pattern
for why this separate class is necesary instead of `DateTimeType | None`
"""


def _fault_tolerant_id_serde(val: Any, serialize: bool = False) -> Any:
    """Needed for the API and safe de-/serialization behavior."""
    try:
        alpha_id = AlphaID(val)
        if serialize:
            return str(alpha_id)
        return alpha_id.formatted
    except Exception:
        return val


IdentifierType: TypeAlias = Annotated[
    Union[MPID, AlphaID],
    BeforeValidator(_fault_tolerant_id_serde),
    PlainSerializer(lambda x: _fault_tolerant_id_serde(x, serialize=True)),
]
"""MPID / AlphaID serde."""


class TypedBandDict(TypedDict):
    """Type def for data stored for cbms or vbms"""

    band_index: dict[str, list[int]]
    kpoint_index: list[int]
    kpoint: KpointType
    energy: float
    projections: dict[str, list[list[float]]]



from typing_extensions import TypedDict

from emmet.core.base import EmmetBaseModel
from emmet.core.types.pymatgen_types.alloy_adapter import AlloyPairType, AlloySystemType


class TypedBoolDict(TypedDict):
    min: bool
    max: bool


class TypedRangeDict(TypedDict):
    min: float
    max: float


class TypedSearchDict(TypedDict):
    alloying_element: list[str]
    band_gap: TypedRangeDict
    energy_above_hull: TypedRangeDict
    formation_energy_per_atom: TypedRangeDict
    formula: list[str]
    id: list[str]
    is_gap_direct: TypedBoolDict
    member_ids: list[str]
    spacegroup_intl_number: TypedRangeDict
    theoretical: TypedBoolDict
    volume_cube_root: TypedRangeDict


class AlloyPairDoc(EmmetBaseModel):
    alloy_pair: AlloyPairType

    pair_id: str

    # fields useful for building search indices
    _search: TypedSearchDict

    @classmethod
    def from_pair(cls, pair):
        return cls(alloy_pair=pair, pair_id=pair.pair_id, _search=pair.search_dict())


class AlloySystemDoc(EmmetBaseModel):
    alloy_system: AlloySystemType

    alloy_id: str

    @classmethod
    def from_pair(cls, alloy_system):
        # Too large to duplicate alloy pairs here.
        alloy_system.alloy_pairs = None  # type: ignore[assignment]
        return cls(alloy_system=alloy_system, alloy_id=alloy_system.alloy_id)



from emmet.core.mpid import (
    MPID,
    MPculeID,
    AlphaID,
    VALID_ALPHA_SEPARATORS,
)
import pytest

import json
from pydantic import BaseModel


def test_mpid():
    assert MPID("mp-3") == MPID("mp-3")
    assert MPID("mp-3") < 3
    assert MPID("mp-3") < MPID("np-3")
    assert MPID("mp-3") > MPID("mp-2")
    assert 3 > MPID("mp-3")
    assert MPID(MPID("mp-1234")) < MPID(1234)
    assert "mp-1234" < MPID(1234)
    assert MPID("1234") > MPID("mp-1234")
    assert MPID("1234") == MPID(1234)
    assert MPID("1234") == "1234"
    assert MPID("mp-12345") > MPID("mp-1234")

    assert min(
        [MPID("mp-44545"), MPID("mp-33"), MPID("mp-2134234"), MPID(33), MPID("33")]
    ) == MPID("mp-33")

    assert (
        len(set([MPID("mp-33"), MPID("mp-44545"), MPID("mp-33"), MPID("mp-2134234")]))
        == 3
    )

    for invalid_value, msg in {
        "-100": "MPID string representation must follow",
        -14: "MPID cannot represent a negative integer.",
    }.items():
        with pytest.raises(ValueError, match=msg):
            MPID(invalid_value)

    MPID(3)
    ulid_mpid = MPID("01HMVV88CCQ6JQ2Y1N8F3ZTVWP-Li")
    assert ulid_mpid.parts == ("01HMVV88CCQ6JQ2Y1N8F3ZTVWP", 0)
    with pytest.raises(ValueError, match="MPID string representation must follow"):
        MPID("GGIRADF")


def test_mpculeid():
    assert MPculeID("b9ba54febc77d2a9177accf4605767db-F6Li1P1-1-2") == MPculeID(
        "b9ba54febc77d2a9177accf4605767db-F6Li1P1-1-2"
    )
    assert (
        MPculeID("b9ba54febc77d2a9177accf4605767db-F6Li1P1-1-2")
        == "b9ba54febc77d2a9177accf4605767db-F6Li1P1-1-2"
    )
    assert MPculeID("b9ba54febc77d2a9177accf4605767db-F6Li1P1-1-2") < MPculeID(
        "b9ba54febc77d2a9177accf4605767db-F6Li1P1-2-1"
    )
    assert MPculeID("b9ba54febc77d2a9177accf4605767db-F6Li1P1-2-1") > MPculeID(
        "b9ba54febc77d2a9177accf4605767db-F6Li1P1-1-2"
    )
    assert MPculeID("mpcule-98bab8f3795eae3fd8e28f5ff2d476e8-C3H8-0-1") < MPculeID(
        "b9ba54febc77d2a9177accf4605767db-F6Li1P1-1-2"
    )


def test_to_str():
    assert str(MPID("mp-149")) == "mp-149"
    assert (
        str(MPculeID("b9ba54febc77d2a9177accf4605767db-F6Li1P1-1-2"))
        == "b9ba54febc77d2a9177accf4605767db-F6Li1P1-1-2"
    )


def test_alpha_id():
    # Test initialization from MPID-like string
    mpid_like_idx = AlphaID("mp-149")
    assert mpid_like_idx == AlphaID(149, prefix="mp", separator="-")
    assert mpid_like_idx._prefix == "mp"
    assert mpid_like_idx._separator == "-"
    assert int(mpid_like_idx) == 149

    assert mpid_like_idx == MPID("mp-149")
    assert mpid_like_idx == AlphaID(MPID("mp-149"))
    assert AlphaID("mvc-160") == MPID("mvc-160")
    assert AlphaID(160) == MPID("160")
    assert mpid_like_idx + MPID("mp-140") == AlphaID("mp-289")
    assert mpid_like_idx - MPID("mp-140") == AlphaID("mp-9")

    # Test relationality - should still compare integers when prefix and separator don't match
    assert mpid_like_idx > MPID("mp-140")
    mvc_idx = AlphaID("mvc-140")
    assert not mpid_like_idx < mvc_idx
    assert mpid_like_idx > mvc_idx

    majik_num = 137

    ref_idx = AlphaID(majik_num, prefix="cats", separator=":")
    assert (
        ref_idx.__repr__()
        == f"AlphaID(cats:{AlphaID._integer_to_alpha_rep(majik_num)})"
    )
    assert ref_idx < AlphaID(
        majik_num + 100, prefix=ref_idx._prefix, separator=ref_idx._separator
    )
    assert ref_idx > AlphaID(
        majik_num - 100, prefix=ref_idx._prefix, separator=ref_idx._separator
    )

    # test padding
    padded_idx = AlphaID(majik_num, prefix="cats", separator=":", padlen=8)
    assert ref_idx == padded_idx
    assert int(ref_idx) == int(padded_idx)
    assert str(padded_idx) == ref_idx._prefix + ref_idx._separator + (
        8 - len(str(ref_idx._identifier))
    ) * "a" + str(ref_idx._identifier)

    # test invalid sep
    for sep in ("^", "&", "%", "$", "#"):
        with pytest.raises(ValueError, match="Invalid separator"):
            AlphaID(majik_num, prefix="v", separator=sep)

    # test copy
    ref_idx_copy = ref_idx.copy()
    ref_idx += 1
    assert ref_idx != ref_idx_copy

    assert AlphaID(majik_num, padlen=10) < AlphaID(majik_num + 1)
    assert AlphaID(majik_num, padlen=10) > AlphaID(majik_num - 1)

    # Test initialization from int

    ordered_seps = list(VALID_ALPHA_SEPARATORS)
    for pfx in (None, "task"):
        for isep, separator in enumerate(ordered_seps):
            idx = AlphaID(majik_num, prefix=pfx, separator=separator)
            assert isinstance(hash(idx), int)

            # Roundtrip
            assert idx == AlphaID(str(idx))

            assert int(idx) == majik_num

            # Test integer equality
            assert idx == majik_num

            # Test in/equality when prefix and separators do not match
            assert idx != AlphaID(majik_num, prefix="unusedprefix")

            # Test equality when separators differ but prefixes are None
            # (separators are unset)
            idx_diff_sep = AlphaID(
                majik_num,
                prefix=pfx,
                separator=ordered_seps[
                    (isep + 1) % len(ordered_seps)
                ],  # ensures separator differs
            )
            if pfx is None:
                assert idx == idx_diff_sep
            else:
                assert idx != idx_diff_sep

            # Test integer addition
            assert idx + 1 == majik_num + 1
            assert idx + 1 == AlphaID(majik_num + 1, prefix=pfx, separator=separator)

            # Test string addition/subtraction
            assert idx + "y" == AlphaID(majik_num + 24, prefix=pfx, separator=separator)
            assert idx - "z" == AlphaID(majik_num - 25, prefix=pfx, separator=separator)

            # Test equality, addition, subtraction of AlphaID
            other_idx = AlphaID("task:100")
            assert other_idx != idx
            if pfx == "task" and separator == ":":
                assert idx - other_idx == majik_num - 100
                assert idx - other_idx == AlphaID(f"task:{majik_num - 100}")

                assert idx + other_idx == majik_num + 100
                assert idx + other_idx == AlphaID(f"task:{majik_num + 100}")
            else:
                match_strs = []
                if pfx != "task":
                    match_strs.append("Prefixes do not match")
                if pfx and separator != ":":
                    match_strs.append("Separators do not match")

                for match_str in match_strs:
                    with pytest.raises(TypeError, match=match_str):
                        idx - other_idx

                    with pytest.raises(TypeError, match=match_str):
                        idx + other_idx

            # Can't add nor compare floats with AlphaID
            with pytest.raises(NotImplementedError, match="Cannot compare AlphaID"):
                idx + 1.0
            with pytest.raises(NotImplementedError, match="Cannot compare AlphaID"):
                idx < float(int(idx))

        # test sorting
        test_ints = (10, 100, 50, 20, 50000)
        test_idxs = [AlphaID(i) for i in test_ints]
        sorted_idxs = sorted(test_idxs)
        assert all(isinstance(_idx, AlphaID) for _idx in sorted_idxs)
        assert [int(v) for v in sorted_idxs] == sorted(test_ints)
        assert max(test_idxs) == AlphaID(max(test_ints))
        assert min(test_idxs) == AlphaID(min(test_ints))

    # test iterative addition / subtraction
    last_val = None
    next_val = None
    for i in range(5000):
        alpha = AlphaID(i)
        assert int(alpha) == i
        if i > 0:
            assert alpha - 1 == last_val
            assert alpha - AlphaID(1) == last_val
            assert alpha == next_val

        last_val = alpha
        next_val = alpha + 1

    # For legacy purposes, ensure that MPIDs with lower values
    # present as int when calling .string
    assert AlphaID("mp-149").string == "mp-149"
    assert AlphaID("mp-3347529").string == "mp-3347529"
    assert AlphaID("mp-3347530").string == str(AlphaID("mp-3347530"))

    for invalid_val, msg in {
        "-100": "Missing prefix and/or identifer.",
        "pfx-": "Missing prefix and/or identifer.",
        ":": "Missing prefix and/or identifer.",
        -100: "AlphaID cannot represent a negative integer.",
    }.items():
        with pytest.raises(ValueError, match=msg):
            AlphaID(invalid_val)

    # Test next safe AlphaID
    # The following words are "obscene" in a
    # non-English language but are safe in English
    # Test works by ensuring that the AlphaID less than
    # a profane word skips the profane word.
    for word in ("pano", "johny", "sega"):
        curr_alpha_id = AlphaID(word)
        prev_alpha_id = curr_alpha_id - 1
        assert prev_alpha_id.next_safe == curr_alpha_id + 1


@pytest.mark.parametrize("id_cls", [MPID, AlphaID])
def test_pydantic(id_cls):
    # test that AlphaID is supported by pydantic de-/serialization

    class TestClass(BaseModel):
        ID: id_cls

    idx = id_cls(101010)
    test_case = TestClass(ID=idx)
    assert test_case.model_dump() == {"ID": idx}
    assert json.loads(test_case.model_dump_json()) == {"ID": str(idx)}

    with pytest.raises(ValueError, match=f"Invalid {id_cls.__name__} Format"):
        TestClass(ID=10.0)



"""Test phonon document models."""

from copy import deepcopy
from tempfile import NamedTemporaryFile

import numpy as np
import pytest
from monty.serialization import loadfn
from pymatgen.core import Structure
from pymatgen.phonon.dos import CompletePhononDos, PhononDos as PmgPhononDos

from emmet.core import ARROW_COMPATIBLE
from emmet.core.phonon import PhononBSDOSDoc, PhononDOS
from emmet.core.testing_utils import assert_schemas_equal

if ARROW_COMPATIBLE:
    import pyarrow as pa

    from emmet.core.arrow import arrowize


@pytest.fixture(scope="module")
def legacy_ph_task(test_dir):
    return loadfn(test_dir / "mp-23302_phonon.json.gz", cls=None)


def test_legacy_migration(legacy_ph_task):
    # ensure that legacy phonon data can be migrated to current schema

    assert all(legacy_ph_task.get(k) for k in ("ph_bs", "ph_dos"))

    # deepcopy needed because migration changes input objects
    ph_doc = PhononBSDOSDoc.migrate_fields(**deepcopy(legacy_ph_task))
    assert_schemas_equal(ph_doc, PhononBSDOSDoc.model_config)

    # check remap of phonon DOS
    for k in ("densities", "frequencies"):
        assert np.all(
            np.abs(
                np.array(getattr(ph_doc.phonon_dos, k, []))
                - np.array(legacy_ph_task["ph_dos"].get(k, []))
            )
            < 1e-6
        )

    # check remap of phonon bandstructure
    assert np.all(
        np.abs(
            np.array(getattr(ph_doc.phonon_bandstructure, "frequencies", []))
            - np.array(legacy_ph_task["ph_bs"].get("bands", []))
        )
        < 1e-6
    )

    # check that Phonon DOS converts to CompletePhononDOS object
    assert isinstance(ph_doc.phonon_dos.to_pmg, CompletePhononDos)
    # when structure or projected DOS fields are missing, `to_pmg` returns a PhononDos object
    for k in (
        "structure",
        "projected_densities",
    ):
        model_config = deepcopy(ph_doc.model_dump())
        model_config["phonon_dos"].pop(k)
        new_task = PhononBSDOSDoc.migrate_fields(**model_config)
        assert isinstance(new_task.phonon_dos.to_pmg, PmgPhononDos)

    temps = [5, 100, 300, 500, 800]
    ref_data = {
        "entropy": [
            0.10690825310989084,
            64.92923008995054,
            118.49759120562243,
            143.87894281334866,
            167.29013558233882,
        ],
        "heat_capacity": [
            0.3096320518171239,
            47.157121712113714,
            49.56965404227773,
            49.77072796011147,
            49.8399369784691,
        ],
        "internal_energy": [
            1988.2321308731728,
            5269.03083181402,
            15060.035206972329,
            24999.080865250606,
            39943.07283219206,
        ],
        "free_energy": [
            1987.697589607624,
            -1223.8921771810346,
            -20489.242154714408,
            -46940.39054142372,
            -93889.03563367906,
        ],
    }
    tprops = ph_doc.compute_thermo_quantities(temps)
    assert all(
        t == pytest.approx(tprops["temperature"][i]) for i, t in enumerate(temps)
    )
    for k, ref_vals in ref_data.items():
        assert all(
            tprops[k][i] == pytest.approx(ref_val) for i, ref_val in enumerate(ref_vals)
        )

    # check population of structure metadata fields
    assert ph_doc.composition == ph_doc.structure.composition
    assert ph_doc.volume == ph_doc.structure.volume
    assert ph_doc.nelements == len(ph_doc.structure.composition.elements)
    assert sorted(ph_doc.elements) == sorted(ph_doc.structure.composition.elements)

    # check that crystal toolkit convenience function works
    assert ph_doc.phonon_bandstructure._to_pmg_es_bs.get_vbm()
    assert ph_doc.phonon_bandstructure._to_pmg_es_bs.get_cbm()


@pytest.mark.skipif(
    not ARROW_COMPATIBLE, reason="pyarrow must be installed to run this test."
)
def test_arrow(tmp_dir, legacy_ph_task):
    ph_doc = PhononBSDOSDoc.migrate_fields(**legacy_ph_task)
    arrow_struct = pa.scalar(
        ph_doc.model_dump(context={"format": "arrow"}), type=arrowize(PhononBSDOSDoc)
    )
    test_arrow_doc = PhononBSDOSDoc(**arrow_struct.as_py(maps_as_pydicts="strict"))
    assert test_arrow_doc

    assert ph_doc.phonon_bandstructure == test_arrow_doc.phonon_bandstructure

    # test primitive structure caching
    assert test_arrow_doc.phonon_bandstructure._primitive_structure is None
    assert isinstance(
        test_arrow_doc.phonon_bandstructure.primitive_structure, Structure
    )
    assert (
        test_arrow_doc.phonon_bandstructure._primitive_structure
        == test_arrow_doc.phonon_bandstructure.primitive_structure
    )


def test_phonopy_dos_integration(tmp_dir):
    """Test phonopy dat-file parsing."""

    temp = 0.063253
    ph_data = np.random.random((20, 2))
    ph_str = f"# Sigma = {temp}\n\t" + "\n\t".join(
        "\t".join(f"{x}" for x in row) for row in ph_data
    )

    temp_file = NamedTemporaryFile(suffix="dat", mode="w")
    temp_file.write(ph_str)
    temp_file.seek(0)
    ph_dos = PhononDOS.from_phonopy(temp_file.name)
    temp_file.close()

    for idx, attr in enumerate(["frequencies", "densities"]):
        assert np.all(np.abs(ph_data[:, idx] - np.array(getattr(ph_dos, attr))) < 1e-6)



from typing import Annotated, TypeVar

from pydantic import BeforeValidator, WrapSerializer
from pymatgen.core.periodic_table import Species
from typing_extensions import NotRequired, TypedDict

MSONableTypedSpeciesDict = TypedDict(
    "MSONableTypedSpeciesDict",
    {
        "@module": str,
        "@class": str,
        "element": str,
        "oxidation_state": float,
        "spin": float,
    },
)


class TypedSpeciesDict(TypedDict):
    element: NotRequired[str | None]
    oxidation_state: NotRequired[float | None]
    spin: NotRequired[float | None]
    occu: NotRequired[float | None]


SpeciesTypeVar = TypeVar("SpeciesTypeVar", Species, TypedSpeciesDict)

SpeciesType = Annotated[
    SpeciesTypeVar,
    BeforeValidator(lambda x: Species.from_dict(x) if isinstance(x, dict) else x),
    WrapSerializer(lambda x, nxt, info: x.as_dict(), return_type=TypedSpeciesDict),
]



import os
import pytest
from tempfile import mkdtemp
from shutil import rmtree

from emmet.core.testing_utils import TEST_FILES_DIR
from emmet.core.types.enums import VaspObject


@pytest.fixture(scope="session")
def test_dir():
    return TEST_FILES_DIR


@pytest.fixture
def tmp_dir():
    """Make temporary directory for tests."""

    old_cwd = os.getcwd()
    new_path = mkdtemp()
    os.chdir(new_path)
    yield
    os.chdir(old_cwd)
    rmtree(new_path)


class SchemaTestData:
    """Dummy class to be used to contain all test data information."""


class SiOptimizeDouble(SchemaTestData):
    folder = "Si_old_double_relax"
    task_files = {
        "relax2": {
            "vasprun_file": "vasprun.xml.relax2.gz",
            "outcar_file": "OUTCAR.relax2.gz",
            "volumetric_files": ["CHGCAR.relax2.gz"],
            "contcar_file": "CONTCAR.relax2.gz",
        },
        "relax1": {
            "vasprun_file": "vasprun.xml.relax1.gz",
            "outcar_file": "OUTCAR.relax1.gz",
            "volumetric_files": ["CHGCAR.relax1.gz"],
            "contcar_file": "CONTCAR.relax1.gz",
        },
    }
    objects = {"relax2": []}
    task_doc = {
        "calcs_reversed": [
            {
                "output": {
                    "vbm": 5.6147,
                    "cbm": 6.2652,
                    "bandgap": 0.6505,
                    "is_gap_direct": False,
                    "is_metal": False,
                    "transition": "(0.000,0.000,0.000)-(0.375,0.375,0.000)",
                    "direct_gap": 2.5561,
                    "run_stats": {
                        "average_memory": 0,
                        "max_memory": 28096.0,
                        "cores": 16,
                    },
                },
                "input": {
                    "incar": {"NSW": 99},
                    "nkpoints": 29,
                    "potcar_spec": [{"titel": "PAW_PBE Si 05Jan2001"}],
                    "structure": {"volume": 40.036816205493494},
                    "is_hubbard": False,
                    "hubbards": None,
                },
            }
        ],
        "analysis": {"delta_volume": 0.8638191769757384, "max_force": 0},
        "input": {
            "structure": {"volume": 40.036816205493494},
            "potcar_spec": [{"titel": "PAW_PBE Si 05Jan2001"}],
            "parameters": {"NSW": 99},
            "is_hubbard": False,
            "hubbards": None,
        },
        "output": {
            "structure": {"volume": 40.90063538246923},
            "energy": -10.84687704,
            "bandgap": 0.6505,
        },
        "custodian": [{"job": {"settings_override": None, "suffix": ".relax1"}}],
        "included_objects": (),
    }


class SiNonSCFUniform(SchemaTestData):

    folder = "Si_uniform"
    task_files = {
        "standard": {
            "vasprun_file": "vasprun.xml.gz",
            "outcar_file": "OUTCAR.gz",
            "volumetric_files": ["CHGCAR.gz"],
            "contcar_file": "CONTCAR.gz",
        }
    }
    objects = {"standard": []}
    task_doc = {
        "calcs_reversed": [
            {
                "output": {
                    "vbm": 5.6162,
                    "cbm": 6.2243,
                    "bandgap": 0.6103,
                    "is_gap_direct": False,
                    "is_metal": False,
                    "transition": "(0.000,0.000,0.000)-(0.000,0.421,0.000)",
                    "direct_gap": 2.5563,
                    "run_stats": {
                        "average_memory": 0,
                        "max_memory": 31004.0,
                        "cores": 16,
                    },
                },
                "input": {
                    "incar": {"NSW": 0},
                    "nkpoints": 220,
                    "potcar_spec": [{"titel": "PAW_PBE Si 05Jan2001"}],
                    "structure": {"volume": 40.88829843008916},
                    "is_hubbard": False,
                    "hubbards": None,
                },
            }
        ],
        "analysis": {"delta_volume": 0, "max_force": 0.5350159115036506},
        "input": {
            "structure": {"volume": 40.88829843008916},
            "potcar_spec": [{"titel": "PAW_PBE Si 05Jan2001"}],
            "parameters": {"NSW": 0},
            "is_hubbard": False,
            "hubbards": None,
        },
        "output": {
            "structure": {"volume": 40.88829843008916},
            "energy": -10.85064059,
            "bandgap": 0.6103,
        },
        "custodian": [{"job": {"settings_override": None, "suffix": ""}}],
        "included_objects": (VaspObject.DOS, VaspObject.BANDSTRUCTURE),
    }


class SiStatic(SchemaTestData):

    folder = "Si_static"
    task_files = {
        "standard": {
            "vasprun_file": "vasprun.xml.gz",
            "outcar_file": "OUTCAR.gz",
            "volumetric_files": ["CHGCAR.gz"],
            "contcar_file": "CONTCAR.gz",
        }
    }
    objects = {"standard": []}
    task_doc = {
        "calcs_reversed": [
            {
                "output": {
                    "vbm": 5.6163,
                    "cbm": 6.2644,
                    "bandgap": 0.6506,
                    "is_gap_direct": False,
                    "is_metal": False,
                    "transition": "(0.000,0.000,0.000)-(0.000,0.375,0.000)",
                    "direct_gap": 2.5563,
                    "run_stats": {
                        "average_memory": 0,
                        "max_memory": 28124.0,
                        "cores": 16,
                    },
                },
                "input": {
                    "incar": {"NSW": 1},
                    "nkpoints": 29,
                    "potcar_spec": [{"titel": "PAW_PBE Si 05Jan2001"}],
                    "structure": {"volume": 40.88829843008916},
                },
            }
        ],
        "analysis": {"delta_volume": 0, "max_force": 0.0},
        "input": {
            "structure": {"volume": 40.88829843008916},
            "potcar_spec": [{"titel": "PAW_PBE Si 05Jan2001"}],
            "parameters": {"NSW": 0},
            "is_hubbard": False,
            "hubbards": None,
        },
        "output": {
            "structure": {"volume": 40.88829843008916},
            "energy": -10.84678256,
            "bandgap": 0.6506,
            "dos_properties": {
                "Si": {
                    "s": {
                        "filling": 0.624669545020562,
                        "center": -2.5151284433409815,
                        "bandwidth": 7.338662205126851,
                        "skewness": 0.6261990748648925,
                        "kurtosis": 2.0074877073276904,
                        "upper_edge": -8.105469079999999,
                    },
                    "p": {
                        "filling": 0.3911927710592045,
                        "center": 3.339269798287516,
                        "bandwidth": 5.999449671419663,
                        "skewness": 0.0173776678056677,
                        "kurtosis": 1.907790411890831,
                        "upper_edge": -0.7536690799999999,
                    },
                }
            },
        },
        "custodian": [{"job": {"settings_override": None, "suffix": ""}}],
        "included_objects": (),
    }


objects = {cls.__name__: cls for cls in SchemaTestData.__subclasses__()}


def get_test_object(object_name):
    """Get the schema test data object from the class name."""
    return objects[object_name]



from typing import TypeVar

from pymatgen.entries.compatibility import (
    Compatibility,
    MaterialsProject2020Compatibility,
    MaterialsProjectAqueousCompatibility,
)
from typing_extensions import TypedDict

TypedCompatibilityDict = TypedDict(
    "TypedCompatibilityDict",
    {
        "@module": str,
        "@class": str,
        "@version": str,
    },
)


class TypedMaterialsProject2020CompatibilityAdapterDict(TypedCompatibilityDict):
    """
    Does not define an as_dict() method, parent(Compatibility) is abstract
    so Monty just serializes to module, class, and version.
    """

    pass


class TypedMaterialsProjectAqueousCompatibilityAdapterDict(TypedCompatibilityDict):
    """
    Does not define an as_dict() method, parent(Compatibility) is abstract
    so Monty just serializes to module, class, and version.
    """

    pass


CompatibilityTypeVar = TypeVar(
    "CompatibilityTypeVar", Compatibility, TypedCompatibilityDict
)

MaterialsProjectAqueousCompatibilityAdapterDictTypeVar = TypeVar(
    "MaterialsProjectAqueousCompatibilityAdapterDictTypeVar",
    MaterialsProjectAqueousCompatibility,
    TypedMaterialsProjectAqueousCompatibilityAdapterDict,
)

MaterialsProject2020CompatibilityAdapterDictTypeVar = TypeVar(
    "MaterialsProject2020CompatibilityAdapterDictTypeVar",
    MaterialsProject2020Compatibility,
    TypedMaterialsProject2020CompatibilityAdapterDict,
)



from __future__ import annotations

from pydantic import BaseModel, Field

from typing import Literal

from emmet.core.types.typing import DateTimeType


class GeneralStoreDoc(BaseModel):
    """
    Defines general store data
    """

    kind: Literal["newsfeed", "seminar", "banner"] | None = Field(
        None, description="Type of the data."
    )

    markdown: str | None = Field(None, description="Markdown data.")

    meta: dict[str, str] | None = Field(None, description="Metadata.")

    last_updated: DateTimeType = Field(
        description="Timestamp for when this document was last updated",
    )



from pathlib import Path

import numpy as np
from emmet.core.openmm.tasks import CalculationOutput


def test_calc_output_from_directory(test_dir):
    output_dir = test_dir / "openmm" / "calc_output"
    # Call the from_directory function
    calc_out = CalculationOutput.from_directory(
        output_dir,
        "state.csv",
        "trajectory.dcd",
        elapsed_time=10.0,
    )

    # Assert the expected attributes of the CalculationOutput object
    assert isinstance(calc_out, CalculationOutput)
    assert calc_out.dir_name == str(output_dir)
    assert calc_out.elapsed_time == 10.0
    assert calc_out.traj_file == "trajectory.dcd"
    assert calc_out.state_file == "state.csv"

    # Assert the contents of the state data
    assert np.array_equal(calc_out.steps_reported[:3], [100, 200, 300])
    assert np.allclose(calc_out.potential_energy[:3], [-26192.4, -25648.6, -25149.6])
    assert np.allclose(calc_out.kinetic_energy[:3], [609.4, 1110.4, 1576.4], atol=0.1)
    assert np.allclose(calc_out.total_energy[:3], [-25583.1, -24538.1, -23573.2])
    assert np.allclose(calc_out.temperature[:3], [29.6, 54.0, 76.6], atol=0.1)
    assert np.allclose(calc_out.volume[:3], [21.9, 21.9, 21.9], atol=0.1)
    assert np.allclose(calc_out.density[:3], [1.0, 0.99, 0.99], atol=0.1)

    # Assert the existence of the DCD and state files
    assert Path(calc_out.dir_name, calc_out.traj_file).exists()
    assert Path(calc_out.dir_name, calc_out.state_file).exists()



"""Generate test for consistency of document models used in API."""

from mp_api.client import MPRester
import json
from importlib_resources import files
from pathlib import Path

test_file = Path(
    files("emmet.core") / ".." / ".." / "tests" / "test_model_fields.py"
).resolve()

document_models = {rester.document_model for rester in MPRester()._all_resters}

model_fields = {}

py_str = '"""Ensure that document models used in API do not change fields."""'
py_str += """
import pytest
from importlib import import_module


"""
for doc_model in sorted(document_models, key=lambda x: f"{x.__module__}.{x.__name__}"):
    model_fields[f"{doc_model.__module__}.{doc_model.__name__}"] = list(
        doc_model.model_fields
    )

py_str += f"ref_model_fields = {json.dumps(model_fields,indent=2)}\n"
py_str += """
@pytest.mark.parametrize('import_str,ref_fields',ref_model_fields.items())
def test_model_field_drift(import_str, ref_fields):
    module, class_name = import_str.rsplit(".",1)
    model_class = getattr(import_module(module),class_name)
    assert set(model_class.model_fields) == set(ref_fields)
"""

test_file.write_text(py_str)



import pytest
from pymatgen.core import Element, Lattice, Structure

from emmet.core.structure import StructureMetadata
from emmet.core.symmetry import CrystalSystem, SymmetryData


@pytest.fixture
def structure():
    test_latt = Lattice.cubic(3.0)
    test_struc = Structure(lattice=test_latt, species=["Fe"], coords=[[0, 0, 0]])
    return test_struc


def test_symmetry(structure):
    symm_doc = SymmetryData.from_structure(structure)

    assert symm_doc.number == 221
    assert symm_doc.point_group == "m-3m"
    assert symm_doc.symbol == "Pm-3m"
    assert symm_doc.crystal_system == CrystalSystem.cubic

    assert symm_doc.model_dump()["crystal_system"] == CrystalSystem.cubic
    assert str(symm_doc.model_dump()["crystal_system"]) == "Cubic"


def test_structure_metadata(structure):
    meta_doc = StructureMetadata.from_structure(structure)

    assert meta_doc.nsites == 1
    assert meta_doc.elements == [Element.Fe]
    assert meta_doc.nelements == 1
    assert meta_doc.formula_pretty == "Fe"
    assert meta_doc.formula_anonymous == "A"
    assert meta_doc.chemsys == "Fe"
    assert meta_doc.volume == pytest.approx(27.0)
    assert meta_doc.density == pytest.approx(3.4345483027509993)
    assert meta_doc.density_atomic == pytest.approx(27.0)


def test_structure_metadata_fewer_fields(structure):
    meta_doc = StructureMetadata.from_structure(
        structure, fields=["nsites", "nelements", "volume"]
    )

    assert meta_doc.nsites == 1
    assert meta_doc.nelements == 1
    assert meta_doc.volume == pytest.approx(27.0)


def test_composition(structure):
    meta_doc = StructureMetadata.from_structure(structure)
    comp_meta_doc = StructureMetadata.from_composition(structure.composition)

    assert meta_doc.elements == comp_meta_doc.elements
    assert meta_doc.nelements == comp_meta_doc.nelements
    assert meta_doc.formula_pretty == comp_meta_doc.formula_pretty
    assert meta_doc.formula_anonymous == comp_meta_doc.formula_anonymous
    assert meta_doc.chemsys == comp_meta_doc.chemsys


def test_schema():
    StructureMetadata.schema()
